# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import collections
import contextlib
import mmap
import os
from pathlib import Path
import re
import subprocess
import tempfile
from typing import NamedTuple

from drgn.helpers.linux.swap import (
    for_each_swap_info,
    swap_file_path,
    swap_is_file,
    swap_total_usage,
    swap_usage_in_pages,
    total_swapcache_pages,
)
from tests.linux_kernel import (
    LinuxKernelTestCase,
    MbrPartition,
    MbrPartitionType,
    fallocate,
    meminfo_field_in_pages,
    mkswap,
    mount,
    skip_unless_have_test_disk,
    swapoff,
    swapon,
    umount,
    write_mbr,
)


class SwapInfo(NamedTuple):
    filename: str
    type: str
    size: int
    used: int
    priority: int


def iter_swaps():
    with open("/proc/swaps", "rb") as f:
        f.readline()
        for line in f:
            tokens = line.split()
            yield SwapInfo(
                filename=tokens[0].decode("unicode-escape"),
                type=tokens[1].decode("ascii"),
                size=int(tokens[2]) * 1024,
                used=int(tokens[3]) * 1024,
                priority=int(tokens[4]),
            )


@contextlib.contextmanager
def tmp_swaps():
    disk = os.environ["DRGN_TEST_DISK"]
    swaps = []

    with contextlib.ExitStack() as exit_stack:
        write_mbr(
            disk,
            [
                MbrPartition(
                    type=MbrPartitionType.LINUX_SWAP,
                    start=1024 * 1024,
                    size=8 * 1024 * 1024,
                ),
                MbrPartition(
                    type=MbrPartitionType.LINUX,
                    start=9 * 1024 * 1024,
                    size=8 * 1024 * 1024,
                ),
            ],
        )

        part_prefix = os.environ["DRGN_TEST_DISK"]
        if part_prefix[:-1].isdigit():
            part_prefix += "p"
        part1 = part_prefix + "1"
        part2 = part_prefix + "2"

        mkswap(part1)
        swapon(part1)
        exit_stack.callback(swapoff, part1)
        swaps.append((part1, False))

        subprocess.check_call(["mke2fs", "-qF", part2])
        tmp = Path(exit_stack.enter_context(tempfile.TemporaryDirectory()))
        mount(part2, tmp, "ext2")
        exit_stack.callback(umount, tmp)
        swap_file = tmp / "swap_file"
        fallocate(swap_file, 0, 1024 * 1024)
        mkswap(swap_file)
        swapon(swap_file)
        exit_stack.callback(swapoff, swap_file)
        swaps.append((swap_file, True))

        yield swaps


@skip_unless_have_test_disk
class TestSwap(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.enterClassContext(tmp_swaps())

    # There is no test_for_each_swap_info() because we test it implicitly in
    # the other tests. These tests also handle duplicate swap filenames, which
    # are unlikely but theoretically possible with namespace shenanigans.

    def test_swap_file_path(self):
        self.assertCountEqual(
            [os.fsdecode(swap_file_path(si)) for si in for_each_swap_info(self.prog)],
            [swap_info.filename for swap_info in iter_swaps()],
        )

    def test_swap_is_file(self):
        self.assertCountEqual(
            [
                (os.fsdecode(swap_file_path(si)), swap_is_file(si))
                for si in for_each_swap_info(self.prog)
            ],
            [
                (swap_info.filename, swap_info.type == "file")
                for swap_info in iter_swaps()
            ],
        )

    def test_swap_usage_in_pages(self):
        actual = collections.Counter()
        for si in for_each_swap_info(self.prog):
            actual[os.fsdecode(swap_file_path(si))] += (
                swap_usage_in_pages(si) * mmap.PAGESIZE
            )

        expected = collections.Counter()
        for swap_info in iter_swaps():
            expected[swap_info.filename] += swap_info.used

        for filename, used in expected.items():
            # Generous delta to allow for stuff getting swapped in/out while
            # the test is running.
            self.assertAlmostEqual(used, actual[filename], delta=1024 * 1024 * 1024)

    def test_swap_total_usage(self):
        swap_usage = swap_total_usage(self.prog)

        with open("/proc/meminfo", "r") as f:
            for line in f:
                if match := re.match(r"Swap(Total|Free):\s*([0-9]+)\s*kB", line):
                    value = int(match.group(2)) * 1024 // mmap.PAGESIZE
                    if match.group(1) == "Total":
                        pages = value
                    else:
                        free_pages = value

        self.assertEqual(swap_usage.pages, pages)
        # Generous delta to allow for stuff getting swapped in/out while the
        # test is running.
        self.assertAlmostEqual(
            swap_usage.free_pages, free_pages, delta=1024 * 1024 * 1024
        )

    def test_total_swapcache_pages(self):
        self.assertAlmostEqual(
            total_swapcache_pages(self.prog),
            meminfo_field_in_pages("SwapCached"),
            delta=1024 * 1024 * 1024,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import re

from _drgn_util.argparseformatter import MultilineHelpFormatter
from tests import TestCase


# We have separate unit tests for multiline_{wrap,fill}(), so just test that
# the integration with argparse works. This is especially important since all
# of HelpFormatter is considered an implementation detail.
class TestMultilineHelpFormatter(TestCase):
    LINES = [
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.",
        "Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.",
    ]

    def test_description(self):
        parser = argparse.ArgumentParser(
            description="\n".join(["|" + line for line in self.LINES]),
            formatter_class=MultilineHelpFormatter,
        )
        regex = re.compile(
            r"\n".join([f" *{re.escape(line)} *" for line in self.LINES]),
            flags=re.MULTILINE,
        )
        self.assertRegex(parser.format_help(), regex)

    def test_argument_help(self):
        parser = argparse.ArgumentParser(formatter_class=MultilineHelpFormatter)
        parser.add_argument(
            "--foo", help="\n".join(["|" + line for line in self.LINES])
        )
        regex = re.compile(
            r"\n".join([f" *{re.escape(line)} *" for line in self.LINES]),
            flags=re.MULTILINE,
        )
        self.assertRegex(parser.format_help(), regex)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

REGISTERS = [
    *[DrgnRegister(f"r{i}") for i in range(32)],
    DrgnRegister("lr"),
    *[DrgnRegister(f"cr{i}") for i in range(8)],
]

# There are two conflicting definitions of DWARF register numbers after 63. The
# original definition appears to be "64-bit PowerPC ELF Application Binary
# Interface Supplement" [1]. The GNU toolchain instead uses its own that was
# later codified in "Power Architecture 64-Bit ELF V2 ABI Specification" [2].
# We use the latter.
#
# 1: https://refspecs.linuxfoundation.org/ELF/ppc64/PPC-elf64abi.html
# 2: https://openpowerfoundation.org/specifications/64bitelfabi/
REGISTER_LAYOUT = [
    DrgnRegisterLayout("lr", size=8, dwarf_number=65),
    *[DrgnRegisterLayout(f"r{i}", size=8, dwarf_number=i) for i in range(32)],
    *[DrgnRegisterLayout(f"cr{i}", size=8, dwarf_number=68 + i) for i in range(8)],
]

STACK_POINTER_REGISTER = "r1"



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from contextlib import redirect_stdout
import io

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from drgn import offsetof, sizeof
from drgn.helpers.common.memory import (
    IdentifiedSymbol,
    identify_address,
    identify_address_all,
    print_annotated_memory,
)
from drgn.helpers.common.stack import print_annotated_stack, print_registers
from drgn.helpers.linux.common import (
    IdentifiedSlabObject,
    IdentifiedTaskStack,
    IdentifiedTaskStruct,
    IdentifiedVmap,
)
from drgn.helpers.linux.mm import pfn_to_virt
from drgn.helpers.linux.sched import idle_task
from tests.linux_kernel import (
    HAVE_FULL_MM_SUPPORT,
    LinuxKernelTestCase,
    online_cpus,
    skip_if_slob,
    skip_unless_have_full_mm_support,
    skip_unless_have_stack_tracing,
    skip_unless_have_test_kmod,
)


class TestIdentifyAddress(LinuxKernelTestCase):
    @skip_unless_have_test_kmod
    def test_identify_symbol(self):
        symbol = self.prog.symbol("drgn_test_function")

        identified = list(identify_address_all(self.prog, symbol.address))
        self.assertIsInstance(identified[0], IdentifiedSymbol)
        self.assertEqual(str(identified[0]), "function symbol: drgn_test_function+0x0")
        # Module symbols are also vmapped.
        self.assertIsInstance(identified[1], IdentifiedVmap)
        self.assertEqual(len(identified), 2)

        self.assertEqual(
            identify_address(self.prog, symbol.address + 1),
            "function symbol: drgn_test_function+0x1",
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_identify_slab_cache(self):
        for size in ("small", "big"):
            with self.subTest(size=size):
                objects = self.prog[f"drgn_test_{size}_slab_objects"]

                if self.prog["drgn_test_slob"]:
                    for obj in objects:
                        if size == "small":
                            self.assertEqual(
                                identify_address(obj), "unknown slab object"
                            )
                        else:
                            self.assertIsNone(identify_address(obj))
                else:
                    for obj in objects:
                        self.assertEqual(
                            identify_address(obj),
                            f"slab object: drgn_test_{size}+0x0",
                        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    @skip_if_slob
    def test_identify_task(self):
        identified = list(identify_address_all(self.prog["drgn_test_kthread"]))
        self.assertIsInstance(identified[0], IdentifiedTaskStruct)
        self.assertEqual(identified[0].task, self.prog["drgn_test_kthread"])
        self.assertEqual(
            str(identified[0]),
            f"task: {self.prog['drgn_test_kthread'].pid.value_()} (drgn_test_kthre)",
        )
        self.assertIsInstance(identified[1], IdentifiedSlabObject)
        self.assertEqual(len(identified), 2)

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    @skip_if_slob
    def test_identify_task_member(self):
        pid_offset = offsetof(self.prog.type("struct task_struct"), "pid")
        self.assertEqual(
            identify_address(self.prog, self.prog["drgn_test_kthread"].pid.address_),
            f"task: {self.prog['drgn_test_kthread'].pid.value_()} (drgn_test_kthre) +{hex(pid_offset)}",
        )

    def test_identify_idle_task_0(self):
        identified = list(identify_address_all(self.prog["init_task"].address_of_()))
        self.assertIsInstance(identified[0], IdentifiedTaskStruct)
        self.assertEqual(identified[0].task, idle_task(self.prog, 0))
        self.assertIsInstance(identified[1], IdentifiedSymbol)
        self.assertGreaterEqual(len(identified), 2)

    @skip_unless_have_full_mm_support
    @skip_if_slob
    def test_identify_idle_task_1(self):
        for cpu in online_cpus():
            if cpu > 0:
                break
        else:
            self.skipTest("online CPU > 0 not found")

        task = idle_task(self.prog, cpu)
        identified = list(identify_address_all(task))
        self.assertIsInstance(identified[0], IdentifiedTaskStruct)
        self.assertEqual(identified[0].task, task)
        self.assertIsInstance(identified[1], IdentifiedSlabObject)
        self.assertEqual(len(identified), 2)

    @skip_unless_have_test_kmod
    def test_identify_vmap(self):
        for cache in (None, {}):
            with self.subTest("uncached" if cache is None else "cached"):
                self.assertTrue(
                    identify_address(
                        self.prog["drgn_test_vmalloc_va"], cache=cache
                    ).startswith("vmap: 0x")
                )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_identify_task_stack(self):
        if self.prog["drgn_test_slob"] and self.prog["drgn_test_slab_stack_enabled"]:
            self.skipTest("test does not support SLOB")

        for cached in (False, True):
            with self.subTest("cached" if cached else "uncached"):
                identified = list(
                    identify_address_all(
                        self.prog,
                        self.prog["drgn_test_kthread"].stack.value_() + 1234,
                        cache={} if cached else None,
                    )
                )
                self.assertIsInstance(identified[0], IdentifiedTaskStack)
                self.assertEqual(
                    str(identified[0]),
                    f"task stack: {self.prog['drgn_test_kthread'].pid.value_()} (drgn_test_kthre) +0x4d2",
                )
                if self.prog["drgn_test_vmap_stack_enabled"]:
                    self.assertIsInstance(identified[1], IdentifiedVmap)
                    self.assertEqual(len(identified), 2)
                elif self.prog["drgn_test_slab_stack_enabled"]:
                    self.assertIsInstance(identified[1], IdentifiedSlabObject)
                    self.assertEqual(len(identified), 2)
                else:
                    self.assertEqual(len(identified), 1)

    @skip_unless_have_full_mm_support
    def test_identify_idle_task_0_stack(self):
        identified = list(identify_address_all(self.prog["init_task"].stack))
        self.assertIsInstance(identified[0], IdentifiedTaskStack)
        self.assertEqual(identified[0].task, idle_task(self.prog, 0))
        # s390x between Linux kernel commits ce3dc447493f ("s390: add support
        # for virtually mapped kernel stacks") (in v4.20) and 944c78376a39
        # ("s390: use init_thread_union aka initial stack for the first
        # process") (in v6.4) allocates init_task.stack.
        if NORMALIZED_MACHINE_NAME == "s390x":
            if self.prog["drgn_test_vmap_stack_enabled"]:
                self.assertIsInstance(identified[1], (IdentifiedSymbol, IdentifiedVmap))
                self.assertEqual(len(identified), 2)
            elif self.prog["drgn_test_slab_stack_enabled"]:
                self.assertIsInstance(
                    identified[1], (IdentifiedSymbol, IdentifiedSlabObject)
                )
                self.assertEqual(len(identified), 2)
            elif len(identified) > 1:
                self.assertIsInstance(identified[1], IdentifiedSymbol)
                self.assertEqual(len(identified), 2)
        else:
            self.assertIsInstance(identified[1], IdentifiedSymbol)
            self.assertGreaterEqual(len(identified), 2)

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_identify_idle_task_1_stack(self):
        if self.prog["drgn_test_slob"] and self.prog["drgn_test_slab_stack_enabled"]:
            self.skipTest("test does not support SLOB")

        for cpu in online_cpus():
            if cpu > 0:
                break
        else:
            self.skipTest("online CPU > 0 not found")

        task = idle_task(self.prog, cpu)

        for cached in (False, True):
            with self.subTest("cached" if cached else "uncached"):
                identified = list(
                    identify_address_all(task.stack, cache={} if cached else None)
                )
                self.assertIsInstance(identified[0], IdentifiedTaskStack)
                self.assertEqual(identified[0].task, task)
                if self.prog["drgn_test_vmap_stack_enabled"]:
                    self.assertIsInstance(identified[1], IdentifiedVmap)
                    self.assertEqual(len(identified), 2)
                elif self.prog["drgn_test_slab_stack_enabled"]:
                    self.assertIsInstance(identified[1], IdentifiedSlabObject)
                    self.assertEqual(len(identified), 2)
                else:
                    self.assertEqual(len(identified), 1)

    @skip_unless_have_test_kmod
    def test_identify_page(self):
        self.assertEqual(
            identify_address(self.prog["drgn_test_page"]),
            f"page: pfn {self.prog['drgn_test_pfn'].value_()}",
        )
        mapping_offset = offsetof(self.prog.type("struct page"), "mapping")
        self.assertEqual(
            identify_address(self.prog["drgn_test_page"].mapping.address_of_()),
            f"page: pfn {self.prog['drgn_test_pfn'].value_()} +{hex(mapping_offset)}",
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_identify_unrecognized(self):
        start_addr = (pfn_to_virt(self.prog["min_low_pfn"])).value_()
        # On s390x, the start address is 0, and identify_address() doesn't
        # allow a negative address.
        if start_addr > 0:
            self.assertIsNone(identify_address(self.prog, start_addr - 1))
        self.assertIsNone(identify_address(self.prog, self.prog["drgn_test_va"]))


class TestPrintAnnotatedMemory(LinuxKernelTestCase):
    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_print_annotated_memory(self):
        f = io.StringIO()
        with redirect_stdout(f):
            print_annotated_memory(
                self.prog,
                self.prog["drgn_test_small_slab_objects"].address_,
                sizeof(self.prog["drgn_test_small_slab_objects"]),
            )
        # For CONFIG_SLOB, we cannot find slab objects. However,
        # print_annotated_memory() should still function with no error. So we
        # don't skip the test here: just skip the assertion.
        if not self.prog["drgn_test_slob"]:
            self.assertIn("slab object: drgn_test_small+0x0", f.getvalue())


class TestPrintAnnotatedStack(LinuxKernelTestCase):
    @skip_unless_have_stack_tracing
    @skip_unless_have_test_kmod
    def test_print_annotated_stack(self):
        trace = self.prog.stack_trace(self.prog["drgn_test_kthread"])

        f = io.StringIO()
        with redirect_stdout(f):
            print_annotated_stack(trace)

        printed_trace = f.getvalue()

        if HAVE_FULL_MM_SUPPORT and not self.prog["drgn_test_slob"]:
            self.assertIn("slab object: drgn_test_small", printed_trace)
        self.assertIn("[function symbol: schedule", printed_trace)
        self.assertIn("schedule at ", printed_trace)


class TestPrintRegisters(LinuxKernelTestCase):
    @skip_unless_have_stack_tracing
    @skip_unless_have_test_kmod
    def test_print_registers(self):
        trace = self.prog.stack_trace(self.prog["drgn_test_kthread_pt_regs"])
        # This is mostly a smoke test: we don't have any well-known registers,
        # and we don't have any specific guarantees about the availability of
        # registers. We can't even assert that all registers are present as hex
        # strings in the output, because some may be printed in decimal, and
        # others may be broken down into smaller fields.
        f = io.StringIO()
        with redirect_stdout(f):
            print_registers(self.prog, trace[0].registers())



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import ctypes
import mmap
import os
import os.path
import re
import tempfile

from drgn import Object
from drgn.helpers.linux.fs import fget
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import CLONE_NEWNS, fork_and_stop, mlock, unshare
from tests.linux_kernel.crash_commands import CrashCommandTestCase


class TestFiles(CrashCommandTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.run_crash_command("set -p")

    def _test_drgn_common(self, cmd, cache=False):
        self.assertIn("for_each_file(", cmd.drgn_option.stdout)

        for variable in (
            "task",
            "pid",
            "command",
            "inode",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

        if cache:
            for variable in ("i_mapping", "nrpages"):
                with self.subTest(variable=variable):
                    self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

            self.assertNotIn("dentry", cmd.drgn_option.globals)
        else:
            for variable in ("file", "dentry"):
                with self.subTest(variable=variable):
                    self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

            self.assertNotIn("i_mapping", cmd.drgn_option.globals)
            self.assertNotIn("nrpages", cmd.drgn_option.globals)

        self.assertIsInstance(cmd.drgn_option.globals["cpu"], int)
        self.assertIsInstance(cmd.drgn_option.globals["fd"], int)
        self.assertIsInstance(cmd.drgn_option.globals["type"], str)
        for variable in (
            "root",
            "cwd",
            "path",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], bytes)

    def test_no_options(self):
        with tempfile.NamedTemporaryFile() as f:
            cmd = self.check_crash_command("files")
            self.assertRegex(cmd.stdout, r"(?m)^ROOT: .* CWD:")
            self.assertRegex(
                cmd.stdout,
                rf"(?m)^\s*{f.fileno()}\s+[0-9a-f]+\s+[0-9a-f]+\s+[0-9a-f]+\s+REG\s+{re.escape(f.name)}",
            )

        self._test_drgn_common(cmd)

    def test_tasks(self):
        with tempfile.NamedTemporaryFile() as f:
            with fork_and_stop() as pid:
                cmd = self.check_crash_command(f"files {os.getpid()} {pid}")
                foreach_cmd = self.check_crash_command(
                    f"foreach {os.getpid()} {pid} files", mode="capture"
                )

            for c in (cmd, foreach_cmd):
                self.assertIn(f"PID: {os.getpid()}", c.stdout)
                self.assertIn(f"PID: {pid}", c.stdout)
                self.assertRegex(c.stdout, r"(?m)^ROOT: .* CWD:")
                self.assertEqual(
                    len(re.findall(rf"^\s*{f.fileno()}\b", c.stdout, flags=re.M)), 2
                )

        self._test_drgn_common(cmd)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_c(self):
        with tempfile.NamedTemporaryFile() as f:
            cmd = self.check_crash_command(f"files -c {os.getpid()}")
            foreach_cmd = self.check_crash_command(
                f"foreach {os.getpid()} files -c", mode="capture"
            )

            for c in (cmd, foreach_cmd):
                self.assertRegex(c.stdout, r"(?m)^ROOT: .* CWD:")
                self.assertRegex(
                    c.stdout,
                    rf"(?m)^\s*{f.fileno()}\s+[0-9a-f]+\s+[0-9a-f]+\s+[0-9]+\s+REG\s+{re.escape(f.name)}",
                )

        self._test_drgn_common(cmd, cache=True)

        self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_d(self):
        with tempfile.NamedTemporaryFile() as f:
            dentry = fget(
                find_task(self.prog, os.getpid()), f.fileno()
            ).f_path.dentry.read_()
            cmd = self.check_crash_command(f"files -d {dentry.value_():x}")

        self.assertRegex(
            cmd.stdout,
            rf"(?m)^\s*{dentry.value_():x}\s+[0-9a-f]+\s+[0-9a-f]+\s+REG\s+.*{re.escape(os.path.basename(f.name))}",
        )

        self.assertEqual(cmd.drgn_option.globals["dentry"], dentry)

        for variable in ("inode", "sb"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self.assertEqual(cmd.drgn_option.globals["type"], "REG")
        self.assertIsInstance(cmd.drgn_option.globals["path"], bytes)

    def test_p(self):
        with tempfile.TemporaryFile(dir="/dev/shm") as f:
            f.write(os.urandom(2 * mmap.PAGESIZE))
            f.flush()
            inode = fget(find_task(self.prog, os.getpid()), f.fileno()).f_inode.read_()
            with mmap.mmap(f.fileno(), 2 * mmap.PAGESIZE) as map:
                f.close()
                address = ctypes.addressof(ctypes.c_char.from_buffer(map))
                # Make sure the pages are faulted in and stay that way.
                mlock(address, 2 * mmap.PAGESIZE)

                cmd = self.check_crash_command(f"files -p {inode.value_():x}")

        self.assertRegex(cmd.stdout, r"(?m)^\s*INODE\s+NRPAGES.*\n\s*[0-9a-f]+\s+2")
        self.assertRegex(
            cmd.stdout, r"(?m)^\s*PAGE\s+PHYSICAL\s+MAPPING\s+INDEX\s+CNT\s+FLAGS.*\n"
        )
        self.assertRegex(cmd.stdout, r"(?m)^\s*[0-9a-f]+\s+[0-9a-f]+\s+[0-9a-f]+\s+\b")

        self.assertIn("inode_for_each_page(", cmd.drgn_option.stdout)
        for variable in (
            "inode",
            "i_mapping",
            "nrpages",
            "page",
            "physical",
            "cnt",
            "flags",
        ):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

        self.assertIsInstance(cmd.drgn_option.globals["index"], int)
        self.assertIsInstance(cmd.drgn_option.globals["decoded_flags"], str)

    def _test_R(self, arg_fn, expect_no_cache, expect_cache, test_foreach=False):
        with tempfile.NamedTemporaryFile() as f:
            arg = arg_fn(f)

            cmd = self.check_crash_command(f"files -R {arg} {os.getpid()}")
            cmds = [cmd]
            if test_foreach:
                foreach_cmd = self.check_crash_command(
                    f"foreach {os.getpid()} files -R {arg}", mode="capture"
                )
                cmds.append(foreach_cmd)

            for c in cmds:
                if expect_no_cache:
                    self.assertRegex(
                        c.stdout,
                        rf"(?m)^\s*{f.fileno()}\s+[0-9a-f]+\s+[0-9a-f]+\s+[0-9a-f]+\s+REG\s+{re.escape(f.name)}",
                    )
                else:
                    self.assertFalse(c.stdout)
            self._test_drgn_common(cmd)
            self.assertIsInstance(cmd.drgn_option.globals["is_match"], bool)
            if test_foreach:
                self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

            cmd = self.check_crash_command(f"files -c -R {arg} {os.getpid()}")
            cmds = [cmd]
            if test_foreach:
                foreach_cmd = self.check_crash_command(
                    f"foreach {os.getpid()} files -c -R {arg}", mode="capture"
                )
                cmds.append(foreach_cmd)

            for c in cmds:
                if expect_cache:
                    self.assertRegex(
                        c.stdout,
                        rf"(?m)^\s*{f.fileno()}\s+[0-9a-f]+\s+[0-9a-f]+\s+[0-9]+\s+REG\s+{re.escape(f.name)}",
                    )
                else:
                    self.assertFalse(c.stdout)
            self._test_drgn_common(cmd, cache=True)
            self.assertIsInstance(cmd.drgn_option.globals["is_match"], bool)
            if test_foreach:
                self.assertEqual(cmd.drgn_option.stdout, foreach_cmd.drgn_option.stdout)

    def test_R_fd(self):
        self._test_R(lambda f: str(f.fileno()), True, True, True)

    def test_R_filename(self):
        self._test_R(lambda f: f.name, True, True)

    def test_R_dentry(self):
        self._test_R(
            lambda f: hex(
                fget(find_task(self.prog, os.getpid()), f.fileno()).f_path.dentry
            ),
            True,
            False,
        )

    def test_R_inode(self):
        self._test_R(
            lambda f: hex(fget(find_task(self.prog, os.getpid()), f.fileno()).f_inode),
            True,
            True,
        )

    def test_R_address_space(self):
        self._test_R(
            lambda f: hex(
                fget(find_task(self.prog, os.getpid()), f.fileno()).f_inode.i_mapping
            ),
            False,
            True,
        )

    def test_R_file(self):
        self._test_R(
            lambda f: hex(fget(find_task(self.prog, os.getpid()), f.fileno())),
            True,
            False,
        )

    def test_R_cwd(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            with fork_and_stop(os.chdir, tmp_dir) as (pid, _):
                cmd = self.check_crash_command(f"files -R {tmp_dir} {pid}")

        self.assertRegex(cmd.stdout, rf"CWD: {re.escape(tmp_dir)}\n$")

        self.assertEqual(cmd.drgn_option.globals["cwd"], os.fsencode(tmp_dir))
        self.assertIsInstance(cmd.drgn_option.globals["is_match"], bool)

    def test_no_files(self):
        cmd = self.check_crash_command(f"files {hex(self.prog['init_task'].address_)}")
        self.assertRegex(cmd.stdout, r"\nNo open files\n$")


class TestMount(CrashCommandTestCase):
    def test_no_options(self):
        self.run_crash_command("set 1")
        cmd = self.check_crash_command("mount")
        self.assertRegex(cmd.stdout, r"(?m)^\s*[0-9a-f]+\s+[0-9a-f]+.*proc")
        self.assertIn("for_each_mount()", cmd.drgn_option.stdout)
        self.assertNotIn("mnt_ns", cmd.drgn_option.globals)

    def test_no_options_in_namespace(self):
        with fork_and_stop(unshare, CLONE_NEWNS) as (pid, _):
            self.run_crash_command(f"set {pid}")
            cmd = self.check_crash_command("mount")
            self.assertIn("for_each_mount(mnt_ns)", cmd.drgn_option.stdout)
            self.assertEqual(
                cmd.drgn_option.globals["mnt_ns"],
                find_task(self.prog, pid).nsproxy.mnt_ns,
            )

    def test_n_pid(self):
        cmd = self.check_crash_command("mount -n 1")
        self.assertRegex(cmd.stdout, r"(?m)^\s*[0-9a-f]+\s+[0-9a-f]+.*proc")
        self.assertIn("for_each_mount(mnt_ns)", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["mnt_ns"], find_task(self.prog, 1).nsproxy.mnt_ns
        )

    def test_n_task(self):
        task = self.prog["init_task"].address_of_()
        cmd = self.check_crash_command(f"mount -n {hex(task)}")
        self.assertRegex(cmd.stdout, r"(?m)^\s*[0-9a-f]+\s+[0-9a-f]+.*proc")
        self.assertIn("for_each_mount(mnt_ns)", cmd.drgn_option.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["mnt_ns"], self.prog["init_task"].nsproxy.mnt_ns
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import ast
from typing import Any, Optional


class NodeVisitor:
    """
    Node visitor based on ast.NodeVisitor that also passes the parent node and
    (right) sibling node.
    """

    def visit(self, node: ast.AST) -> Any:
        return self._visit(node, None, None)

    def _visit(
        self, node: ast.AST, parent: Optional[ast.AST], sibling: Optional[ast.AST]
    ) -> Any:
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, None)
        if visitor is None:
            self.generic_visit(node)
        else:
            return visitor(node, parent, sibling)

    def generic_visit(self, node: ast.AST) -> None:
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                prev = None
                for item in value:
                    if isinstance(item, ast.AST):
                        if prev:
                            self._visit(prev, node, item)
                        prev = item
                if prev:
                    self._visit(prev, node, None)
            elif isinstance(value, ast.AST):
                self._visit(value, node, None)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import io
import os
import sys
import types

from drgn.commands._crash.common import CRASH_COMMAND_NAMESPACE
from tests import with_default_prog
from tests.linux_kernel import LinuxKernelTestCase


class CrashCommandTestCaseBase:
    """A mix-in to add to a general test case class for testing crash commands"""

    # Run a crash command and capture its stdout and stderr.
    @classmethod
    def run_crash_command(cls, command):
        with contextlib.redirect_stdout(
            io.StringIO()
        ) as stdout, contextlib.redirect_stderr(io.StringIO()) as stderr:
            CRASH_COMMAND_NAMESPACE.run(cls.prog, command)
        return types.SimpleNamespace(stdout=stdout.getvalue(), stderr=stderr.getvalue())

    # Run a crash command with --drgn. Capture its stdout and check that it
    # doesn't write anything to stderr.
    #
    # mode must be "capture", "compile", or "exec".
    #
    # If mode is "compile" or "exec", check that it outputs valid, non-empty
    # Python code.
    #
    # If mode is "exec", also execute the code and capture any globals it sets.
    def run_crash_command_drgn_option(self, command, mode="exec"):
        assert mode in {"capture", "compile", "exec"}

        ret = self.run_crash_command(command + " --drgn")

        self.assertFalse(ret.stderr)

        if os.getenv("DRGN_TEST_LOG_CRASH_DRGN"):
            sys.stderr.write(
                f"""
{'=' * 88}
%crash {command} --drgn
{'-' * 88}
{ret.stdout}\
{'=' * 88}
"""
            )
            sys.stderr.flush()

        if mode == "compile" or mode == "exec":
            self.assertTrue(ret.stdout)
            ret.code = compile(ret.stdout, command + " --drgn", "exec")

        if mode == "exec":
            ret.globals = {"prog": self.prog}
            with with_default_prog(self.prog):
                exec(ret.stdout, ret.globals)

        return ret

    # Run a crash command with and without --drgn. mode is passed to
    # run_crash_command_drgn_option().
    def check_crash_command(self, command, mode="exec"):
        drgn_option = self.run_crash_command_drgn_option(command, mode)

        ret = self.run_crash_command(command)

        if os.getenv("DRGN_TEST_LOG_CRASH_OUTPUT"):
            sys.stderr.write(
                f"""
{'=' * 88}
%crash {command}
{'-' * 88}
{ret.stdout}\
{'=' * 88}
"""
            )
            sys.stderr.flush()

        ret.drgn_option = drgn_option
        return ret


class CrashCommandTestCase(CrashCommandTestCaseBase, LinuxKernelTestCase):
    pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import ast
from typing import Optional


class CodeGenError(Exception):
    def __init__(
        self,
        message: str,
        filename: str,
        lineno: Optional[int] = None,
        columnno: Optional[int] = None,
    ) -> None:
        self.message = message
        self.filename = filename
        self.lineno = lineno
        self.columnno = columnno

    def __str__(self) -> str:
        parts = [self.filename]
        if self.lineno is not None:
            parts.append(str(self.lineno))
            if self.columnno is not None:
                parts.append(str(self.columnno))
        return f"{':'.join(parts)}: error: {self.message}"


_C_ESCAPE_CHAR = []
for c in range(256):
    if c == 0:
        e = r"\0"
    elif c == 7:
        e = r"\a"
    elif c == 8:
        e = r"\b"
    elif c == 9:
        e = r"\t"
    elif c == 10:
        e = r"\n"
    elif c == 11:
        e = r"\v"
    elif c == 12:
        e = r"\f"
    elif c == 13:
        e = r"\r"
    elif c == 34:
        e = r"\""
    elif c == 39:
        e = r"\'"
    elif c == 92:
        e = r"\\"
    elif 32 <= c <= 126:
        e = chr(c)
    else:
        e = f"\\x{c:02x}"
    _C_ESCAPE_CHAR.append(e)


def c_char_ord_literal(o: int) -> str:
    """Return a C character literal for a Unicode code point."""
    if o == 34:  # ord('"')
        return "'\"'"
    elif o <= 0xFF:
        return "'" + _C_ESCAPE_CHAR[o] + "'"
    elif o <= 0xFFFF:
        return f"'\\u{o:04x}'"
    else:
        return f"'\\u{o:08x}'"


def c_bytes_literal(s: bytes) -> str:
    """Return a C string literal for a byte string."""
    result = ['"']
    for c in s:
        if c == 39:  # ord("'")
            result.append("'")
        else:
            result.append(_C_ESCAPE_CHAR[c])
    result.append('"')
    return "".join(result)


def c_string_literal(s: str) -> str:
    """Return a C string literal for a string."""
    result = ['"']
    for c in s:
        o = ord(c)
        if o == 39:  # ord("'")
            result.append("'")
        elif o <= 0xFF:
            result.append(_C_ESCAPE_CHAR[o])
        elif o <= 0xFFFF:
            result.append(f"\\u{o:04x}")
        else:
            result.append(f"\\u{o:08x}")
    result.append('"')
    return "".join(result)


def parse_c_string_literal(s: str) -> str:
    # Python string literals are close enough to C string literals for now.
    return ast.literal_eval(s)  # type: ignore[no-any-return]



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import NULL, Object
from drgn.helpers.linux.radixtree import radix_tree_for_each, radix_tree_lookup
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestRadixTree(LinuxKernelTestCase):
    def test_radix_tree_lookup_empty(self):
        root = self.prog["drgn_test_radix_tree_empty"].address_of_()
        self.assertIdentical(radix_tree_lookup(root, 0), NULL(self.prog, "void *"))
        self.assertIdentical(radix_tree_lookup(root, 100000), NULL(self.prog, "void *"))

    def test_radix_tree_for_each_empty(self):
        root = self.prog["drgn_test_radix_tree_empty"].address_of_()
        self.assertIdentical(list(radix_tree_for_each(root)), [])

    def test_radix_tree_lookup_one(self):
        root = self.prog["drgn_test_radix_tree_one"].address_of_()
        self.assertIdentical(radix_tree_lookup(root, 0), NULL(self.prog, "void *"))
        self.assertIdentical(radix_tree_lookup(root, 665), NULL(self.prog, "void *"))
        self.assertIdentical(
            radix_tree_lookup(root, 666), Object(self.prog, "void *", 0xDEADB00)
        )
        self.assertIdentical(radix_tree_lookup(root, 667), NULL(self.prog, "void *"))
        self.assertIdentical(radix_tree_lookup(root, 100000), NULL(self.prog, "void *"))

    def test_radix_tree_for_each_one(self):
        root = self.prog["drgn_test_radix_tree_one"].address_of_()
        self.assertIdentical(
            list(radix_tree_for_each(root)),
            [(666, Object(self.prog, "void *", 0xDEADB00))],
        )

    def test_radix_tree_lookup_one_at_zero(self):
        root = self.prog["drgn_test_radix_tree_one_at_zero"].address_of_()
        self.assertIdentical(
            radix_tree_lookup(root, 0), Object(self.prog, "void *", 0x1234)
        )
        self.assertIdentical(radix_tree_lookup(root, 1), NULL(self.prog, "void *"))
        self.assertIdentical(radix_tree_lookup(root, 100000), NULL(self.prog, "void *"))

    def test_radix_tree_for_each_one_at_zero(self):
        root = self.prog["drgn_test_radix_tree_one_at_zero"].address_of_()
        self.assertIdentical(
            list(radix_tree_for_each(root)), [(0, Object(self.prog, "void *", 0x1234))]
        )

    def test_radix_tree_lookup_sparse(self):
        root = self.prog["drgn_test_radix_tree_sparse"].address_of_()
        self.assertIdentical(radix_tree_lookup(root, 0), NULL(self.prog, "void *"))
        self.assertIdentical(
            radix_tree_lookup(root, 1), Object(self.prog, "void *", 0x1234)
        )
        self.assertIdentical(radix_tree_lookup(root, 2), NULL(self.prog, "void *"))
        self.assertIdentical(
            radix_tree_lookup(root, 0x40000000), NULL(self.prog, "void *")
        )
        self.assertIdentical(
            radix_tree_lookup(root, 0x80000000), NULL(self.prog, "void *")
        )
        self.assertIdentical(
            radix_tree_lookup(root, 0x80800000), NULL(self.prog, "void *")
        )
        self.assertIdentical(
            radix_tree_lookup(root, 0x80808000), NULL(self.prog, "void *")
        )
        self.assertIdentical(
            radix_tree_lookup(root, 0x80808080), Object(self.prog, "void *", 0x5678)
        )
        self.assertIdentical(
            radix_tree_lookup(root, 0xFFFFFFFE), NULL(self.prog, "void *")
        )
        self.assertIdentical(
            radix_tree_lookup(root, 0xFFFFFFFF), Object(self.prog, "void *", 0x9ABC)
        )

    def test_radix_tree_for_each_sparse(self):
        root = self.prog["drgn_test_radix_tree_sparse"].address_of_()
        self.assertIdentical(
            list(radix_tree_for_each(root)),
            [
                (1, Object(self.prog, "void *", 0x1234)),
                (0x80808080, Object(self.prog, "void *", 0x5678)),
                (0xFFFFFFFF, Object(self.prog, "void *", 0x9ABC)),
            ],
        )

    def test_radix_tree_lookup_multi_index(self):
        try:
            root = self.prog["drgn_test_radix_tree_multi_order"].address_of_()
        except KeyError:
            # Radix tree multi-order support only exists between Linux kernel
            # commits e61452365372 ("radix_tree: add support for multi-order
            # entries") (in v4.6) and 3a08cd52c37c ("radix tree: Remove
            # multiorder support") (in v4.20), and only if
            # CONFIG_RADIX_TREE_MULTIORDER=y.
            self.skipTest("kernel does not have multi-order radix trees")
        self.assertIdentical(
            radix_tree_lookup(root, 0x80807FFF), NULL(self.prog, "void *")
        )
        for index in range(0x80808000, 0x80808200):
            with self.subTest(index=index):
                self.assertIdentical(
                    radix_tree_lookup(root, index), Object(self.prog, "void *", 0x1234)
                )
        self.assertIdentical(
            radix_tree_lookup(root, 0x80808200), NULL(self.prog, "void *")
        )

    def test_radix_tree_for_each_multi_index(self):
        try:
            root = self.prog["drgn_test_radix_tree_multi_order"].address_of_()
        except KeyError:
            # See test_radix_tree_lookup_multi_index().
            self.skipTest("kernel does not have multi-order radix trees")
        self.assertIdentical(
            list(radix_tree_for_each(root)),
            [(0x80808000, Object(self.prog, "void *", 0x1234))],
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import filename_matches
from tests import TestCase


class TestFilenameMatches(TestCase):
    def test_filename_matches(self):
        self.assertTrue(filename_matches("ab/cd/ef", "ef"))
        self.assertTrue(filename_matches("ab/cd/ef", "cd/ef"))
        self.assertFalse(filename_matches("ab/cd/ef", "d/ef"))
        self.assertFalse(filename_matches("/ef", "cd/ef"))
        self.assertTrue(filename_matches("/abc", "abc"))
        self.assertFalse(filename_matches("abc", "/abc"))

    def test_empty(self):
        self.assertTrue(filename_matches("", ""))
        self.assertTrue(filename_matches("ab", ""))
        self.assertFalse(filename_matches("", "ab"))

    def test_one_component(self):
        self.assertTrue(filename_matches("ab", "ab"))
        self.assertFalse(filename_matches("ab", "cd"))

    def test_multiple_components(self):
        self.assertTrue(filename_matches("ab/cd/ef", "ef"))
        self.assertTrue(filename_matches("ab/cd/ef", "cd/ef"))
        self.assertFalse(filename_matches("ab/cd/ef", "cd"))
        self.assertFalse(filename_matches("ab/cd/ef", "ab/ef"))
        self.assertFalse(filename_matches("ef", "ab/cd/ef"))

    def test_component_substring(self):
        self.assertFalse(filename_matches("ab/cd/ef", "d/ef"))

    def test_absolute(self):
        self.assertTrue(filename_matches("/abc", "abc"))
        self.assertFalse(filename_matches("abc", "/abc"))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import os
from pathlib import Path
import resource
import signal
from threading import Condition, Thread
import time
import unittest

from drgn import container_of
from drgn.helpers.linux.cgroup import cgroup_get_from_path
from drgn.helpers.linux.cpumask import for_each_possible_cpu
from drgn.helpers.linux.pid import find_task
from drgn.helpers.linux.sched import (
    cfs_rq_for_each_entity,
    cpu_curr,
    cpu_rq,
    get_task_state,
    idle_task,
    loadavg,
    rq_for_each_fair_task,
    rq_for_each_rt_task,
    sched_entity_is_task,
    sched_entity_to_task,
    task_cpu,
    task_group_name,
    task_on_cpu,
    task_since_last_arrival_ns,
    task_state_to_char,
    task_thread_info,
    thread_group_leader,
)
from tests.linux_kernel import (
    LinuxKernelTestCase,
    fork_and_stop,
    proc_state,
    skip_unless_have_test_kmod,
    wait_until,
)
from tests.linux_kernel.helpers.test_cgroup import tmp_cgroup


@contextlib.contextmanager
def tmp_cgroup_with_cpu_controller():
    with tmp_cgroup() as cgroup_dir:
        if "cpu" not in (cgroup_dir / "cgroup.controllers").read_text().split():
            raise unittest.SkipTest("cgroup CPU controller not enabled")
        yield cgroup_dir


class TestSched(LinuxKernelTestCase):
    @skip_unless_have_test_kmod
    def test_task_thread_info(self):
        self.assertEqual(
            task_thread_info(self.prog["drgn_test_kthread"]),
            self.prog["drgn_test_kthread_info"],
        )

    def test_task_cpu(self):
        cpu = os.cpu_count() - 1
        with fork_and_stop(os.sched_setaffinity, 0, (cpu,)) as (pid, _):
            self.assertEqual(task_cpu(find_task(self.prog, pid)), cpu)

    def test_task_state_to_char(self):
        task = find_task(self.prog, os.getpid())
        self.assertEqual(task_state_to_char(task), "R")
        self.assertEqual(get_task_state(task), "R (running)")

        pid = os.fork()
        try:
            if pid == 0:
                try:
                    while True:
                        signal.pause()
                finally:
                    os._exit(1)

            task = find_task(self.prog, pid)

            wait_until(lambda: proc_state(pid) == "S")
            self.assertEqual(task_state_to_char(task), "S")
            self.assertEqual(get_task_state(task), "S (sleeping)")

            os.kill(pid, signal.SIGSTOP)
            wait_until(lambda: proc_state(pid) == "T")
            self.assertEqual(task_state_to_char(task), "T")

            os.kill(pid, signal.SIGKILL)
            wait_until(lambda: proc_state(pid) == "Z")
            self.assertEqual(task_state_to_char(task), "Z")
        finally:
            os.kill(pid, signal.SIGKILL)
            os.waitpid(pid, 0)

    def test_task_on_cpu(self):
        task = find_task(self.prog, os.getpid())
        self.assertTrue(task_on_cpu(task))

        with fork_and_stop() as pid:
            task = find_task(self.prog, pid)
            self.assertFalse(task_on_cpu(task))

    def test_cpu_curr(self):
        task = find_task(self.prog, os.getpid())
        cpu = os.cpu_count() - 1
        old_affinity = os.sched_getaffinity(0)
        os.sched_setaffinity(0, (cpu,))
        try:
            self.assertEqual(cpu_curr(self.prog, cpu), task)
        finally:
            os.sched_setaffinity(0, old_affinity)

    def test_idle_task(self):
        for cpu in for_each_possible_cpu(self.prog):
            task = idle_task(self.prog, cpu)
            if cpu == 0:
                self.assertEqual(task, self.prog["init_task"].address_of_())
            else:
                self.assertEqual(task.comm.string_(), f"swapper/{cpu}".encode())

    def test_loadavg(self):
        values = loadavg(self.prog)
        self.assertEqual(len(values), 3)
        self.assertTrue(all(v >= 0.0 for v in values))

    def test_task_since_last_arrival_ns(self):
        with fork_and_stop() as pid:
            time.sleep(0.01)
            # Forcing the process to migrate also forces the rq clock to update
            # so we can get a reliable reading.
            affinity = os.sched_getaffinity(pid)
            if len(affinity) > 1:
                other_affinity = {affinity.pop()}
                os.sched_setaffinity(pid, affinity)
                os.sched_setaffinity(pid, other_affinity)
            task = find_task(self.prog, pid)
            self.assertGreaterEqual(task_since_last_arrival_ns(task), 10000000)

    def test_rq_for_each_fair_task(self):
        old_affinity = os.sched_getaffinity(0)
        cpu = max(old_affinity)
        os.sched_setaffinity(0, (cpu,))
        self.addCleanup(os.sched_setaffinity, 0, old_affinity)

        rq = cpu_rq(self.prog, cpu)
        pids = [task.pid.value_() for task in rq_for_each_fair_task(rq)]
        self.assertIn(os.getpid(), pids)

    def test_rq_for_each_rt_task(self):
        old_affinity = os.sched_getaffinity(0)
        cpu = max(old_affinity)
        os.sched_setaffinity(0, (cpu,))
        self.addCleanup(os.sched_setaffinity, 0, old_affinity)

        old_rlimit = resource.getrlimit(resource.RLIMIT_RTTIME)
        resource.setrlimit(
            resource.RLIMIT_RTTIME, (resource.RLIM_INFINITY, resource.RLIM_INFINITY)
        )
        self.addCleanup(resource.setrlimit, resource.RLIMIT_RTTIME, old_rlimit)

        old_scheduler = os.sched_getscheduler(0)
        old_param = os.sched_getparam(0)
        os.sched_setscheduler(0, os.SCHED_RR, os.sched_param(1))
        self.addCleanup(os.sched_setscheduler, 0, old_scheduler, old_param)

        rq = cpu_rq(self.prog, cpu)
        pids = [task.pid.value_() for task in rq_for_each_rt_task(rq)]
        self.assertIn(os.getpid(), pids)

    def test_sched_entity_is_task(self):
        task = find_task(self.prog, os.getpid())
        self.assertTrue(sched_entity_is_task(task.se.address_of_()))

    def test_sched_entity_is_task_false(self):
        with tmp_cgroup_with_cpu_controller() as cgroup_dir:
            cgrp = cgroup_get_from_path(self.prog, cgroup_dir.name)
            css = cgrp.subsys[self.prog["cpu_cgrp_id"]]
            task_group = container_of(css, "struct task_group", "css")
            cpu = min(os.sched_getaffinity(0))
            se = task_group.se[cpu].read_()
            self.assertFalse(sched_entity_is_task(se))

    def test_sched_entity_to_task(self):
        task = find_task(self.prog, os.getpid())
        self.assertEqual(sched_entity_to_task(task.se.address_of_()), task)

    def test_task_group_name(self):
        with tmp_cgroup_with_cpu_controller() as cgroup_dir:
            cgrp = cgroup_get_from_path(self.prog, cgroup_dir.name)
            css = cgrp.subsys[self.prog["cpu_cgrp_id"]]
            task_group = container_of(css, "struct task_group", "css")
            self.assertEqual(task_group_name(task_group), cgroup_dir.name.encode())

    def test_cfs_rq_for_each_entity(self):
        old_affinity = os.sched_getaffinity(0)
        cpu = max(old_affinity)
        self.addCleanup(os.sched_setaffinity, 0, old_affinity)
        os.sched_setaffinity(0, (cpu,))

        with tmp_cgroup_with_cpu_controller() as cgroup_dir:
            old_cgroup = (
                Path("/proc/self/cgroup").read_text().strip().partition("::")[2]
            )
            old_cgroup_procs = (
                cgroup_dir.parent / old_cgroup.lstrip("/") / "cgroup.procs"
            )
            (cgroup_dir / "cgroup.procs").write_text(str(os.getpid()))
            try:
                cgrp = cgroup_get_from_path(self.prog, cgroup_dir.name)
                css = cgrp.subsys[self.prog["cpu_cgrp_id"]]
                task_group = container_of(css, "struct task_group", "css")
                task_group_se = task_group.se[cpu].read_()

                task = find_task(self.prog, os.getpid())
                task_se = task.se.address_of_()

                entities = list(
                    cfs_rq_for_each_entity(cpu_rq(self.prog, cpu).cfs.address_of_())
                )
                self.assertIn((task_group_se, 0, True, False), entities)
                self.assertIn((task_se, 1, True, True), entities)
            finally:
                old_cgroup_procs.write_text(str(os.getpid()))

    def test_thread_group_leader(self):
        condition = Condition()

        def thread_func():
            with condition:
                condition.wait()

        thread = Thread(target=thread_func)
        try:
            thread.start()

            self.assertTrue(thread_group_leader(find_task(self.prog, os.getpid())))
            self.assertFalse(
                thread_group_leader(find_task(self.prog, thread.native_id))
            )
        finally:
            with condition:
                condition.notify_all()
            thread.join()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from pathlib import Path
import time

from drgn import cast
from drgn.helpers.linux.timekeeping import (
    ktime_get_boottime_seconds,
    ktime_get_clocktai_seconds,
    ktime_get_coarse_boottime_ns,
    ktime_get_coarse_clocktai_ns,
    ktime_get_coarse_ns,
    ktime_get_coarse_real_ns,
    ktime_get_real_seconds,
    ktime_get_seconds,
    ktime_to_ns,
    uptime,
    uptime_pretty,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod

# These aren't available in the time module as of Python 3.14.
CLOCK_REALTIME_COARSE = getattr(time, "CLOCK_REALTIME_COARSE", 5)
CLOCK_MONOTONIC_COARSE = getattr(time, "CLOCK_MONOTONIC_COARSE", 6)


class TestTimekeeping(LinuxKernelTestCase):
    def assert_in_range(self, a, b, c):
        self.assertTrue(a <= b <= c, f"{b} is not in range [{a}, {c}]")

    @skip_unless_have_test_kmod
    def test_ktime_to_ns(self):
        self.assertEqual(
            ktime_to_ns(self.prog["drgn_test_ktime"]), self.prog["drgn_test_ktime_ns"]
        )

    def test_ktime_get_seconds(self):
        t1 = int(time.clock_gettime(CLOCK_MONOTONIC_COARSE))
        t2 = ktime_get_seconds(self.prog)
        t3 = int(time.clock_gettime(CLOCK_MONOTONIC_COARSE))

        self.assert_in_range(t1, t2.value_(), t3)
        self.assertIdentical(t2, cast("time64_t", t2))

    def test_ktime_get_coarse_ns(self):
        t1 = time.clock_gettime_ns(CLOCK_MONOTONIC_COARSE)
        t2 = ktime_get_coarse_ns(self.prog)
        t3 = time.clock_gettime_ns(CLOCK_MONOTONIC_COARSE)

        self.assert_in_range(t1, t2.value_(), t3)
        self.assertIdentical(t2, cast("u64", t2))

    def test_ktime_get_real_seconds(self):
        t1 = int(time.clock_gettime(CLOCK_REALTIME_COARSE))
        t2 = ktime_get_real_seconds(self.prog)
        t3 = int(time.clock_gettime(CLOCK_REALTIME_COARSE))

        self.assert_in_range(t1, t2.value_(), t3)
        self.assertIdentical(t2, cast("time64_t", t2))

    def test_ktime_get_coarse_real_ns(self):
        t1 = time.clock_gettime_ns(CLOCK_REALTIME_COARSE)
        t2 = ktime_get_coarse_real_ns(self.prog)
        t3 = time.clock_gettime_ns(CLOCK_REALTIME_COARSE)

        self.assert_in_range(t1, t2.value_(), t3)
        self.assertIdentical(t2, cast("u64", t2))

    @skip_unless_have_test_kmod
    def test_ktime_get_boottime_seconds(self):
        # There is no CLOCK_BOOTTIME_COARSE, so the test module exposes it in
        # sysfs.
        path = Path("/sys/kernel/drgn_test/boottime_seconds")

        t1 = int(path.read_text())
        t2 = ktime_get_boottime_seconds(self.prog)
        t3 = int(path.read_text())

        self.assert_in_range(t1, t2.value_(), t3)
        self.assertIdentical(t2, cast("time64_t", t2))

    @skip_unless_have_test_kmod
    def test_ktime_get_coarse_boottime_ns(self):
        # There is no CLOCK_BOOTTIME_COARSE, so the test module exposes it in
        # sysfs.
        path = Path("/sys/kernel/drgn_test/coarse_boottime_ns")

        t1 = int(path.read_text())
        t2 = ktime_get_coarse_boottime_ns(self.prog)
        t3 = int(path.read_text())

        self.assert_in_range(t1, t2.value_(), t3)
        self.assertIdentical(t2, cast("u64", t2))

    @skip_unless_have_test_kmod
    def test_ktime_get_clocktai_seconds(self):
        # There is no CLOCK_TAI_COARSE, so the test module exposes it in sysfs.
        path = Path("/sys/kernel/drgn_test/clocktai_seconds")

        t1 = int(path.read_text())
        t2 = ktime_get_clocktai_seconds(self.prog)
        t3 = int(path.read_text())

        self.assert_in_range(t1, t2.value_(), t3)
        self.assertIdentical(t2, cast("time64_t", t2))

    @skip_unless_have_test_kmod
    def test_ktime_get_coarse_clocktai_ns(self):
        # There is no CLOCK_TAI_COARSE, so the test module exposes it in sysfs.
        path = Path("/sys/kernel/drgn_test/coarse_clocktai_ns")

        t1 = int(path.read_text())
        t2 = ktime_get_coarse_clocktai_ns(self.prog)
        t3 = int(path.read_text())

        self.assert_in_range(t1, t2.value_(), t3)
        self.assertIdentical(t2, cast("u64", t2))

    @skip_unless_have_test_kmod
    def test_uptime(self):
        # There is no CLOCK_BOOTTIME_COARSE, so the test module exposes it in
        # sysfs.
        path = Path("/sys/kernel/drgn_test/coarse_boottime_ns")

        t1 = int(path.read_text())
        t2 = uptime(self.prog)
        t3 = int(path.read_text())

        self.assert_in_range(t1 / 1e9, t2, t3 / 1e9)
        self.assertIsInstance(t2, float)

    def test_uptime_pretty(self):
        # Just test that it succeeds.
        uptime_pretty(self.prog)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os.path
from typing import Any, Dict, NamedTuple, Optional, Sequence, Union
import zlib

from _drgn_util.elf import ET, SHF, SHT
from tests.assembler import _append_sleb128, _append_uleb128
from tests.dwarf import DW_AT, DW_FORM, DW_LNCT, DW_LNE, DW_LNS, DW_TAG, DW_UT
from tests.elfwriter import ElfSection, create_elf_file


class DwarfAttrib(NamedTuple):
    name: DW_AT
    form: DW_FORM
    value: Any


class DwarfLabel(NamedTuple):
    name: str


class DwarfDie(NamedTuple):
    tag: DW_TAG
    attribs: Sequence[DwarfAttrib] = ()
    children: Sequence[Union["DwarfDie", DwarfLabel]] = ()


class DwarfLnp(NamedTuple):
    instructions: Sequence[Sequence[Any]]
    comp_dir: str
    file_names: Sequence[str]
    default_is_stmt: bool = True


class DwarfUnit(NamedTuple):
    type: DW_UT
    die: DwarfDie
    die_label: Optional[str] = None
    dwo_id: Optional[int] = None
    type_signature: Optional[int] = None
    type_offset: Optional[str] = None
    lnp: Optional[DwarfLnp] = None


def create_dwarf_lnp(rows, comp_dir: str, file_name: str) -> DwarfLnp:
    instructions = []
    file_names = {file_name: 1}
    current_pc = 0
    current_file = 1
    current_line = 1
    current_column = 0
    for pc, file_name, line, column in rows:
        if pc > current_pc:
            instructions.append((DW_LNS.advance_pc, pc - current_pc))
        else:
            assert pc == 0 and current_pc == 0

        if file_name is None:
            assert line is None
            assert column is None
            instructions.append((DW_LNE.end_sequence,))
        else:
            file = file_names.setdefault(file_name, len(file_names) + 1)
            if file != current_file:
                instructions.append((DW_LNS.set_file, file))
            if line != current_line:
                instructions.append((DW_LNS.advance_line, line - current_line))
            if column != current_column:
                instructions.append((DW_LNS.set_column, column))
            instructions.append((DW_LNS.copy,))
        current_pc = pc
        current_file = file
        current_line = line
        current_column = column
    assert instructions[-1] == (DW_LNE.end_sequence,)

    return DwarfLnp(
        instructions=instructions, comp_dir=comp_dir, file_names=list(file_names)
    )


def _compile_debug_abbrev(units, use_dw_form_indirect):
    buf = bytearray()
    code = 1

    def aux(die):
        if isinstance(die, DwarfLabel):
            return
        nonlocal code
        _append_uleb128(buf, code)
        code += 1
        _append_uleb128(buf, die.tag)
        buf.append(bool(die.children))
        for attrib in die.attribs:
            _append_uleb128(buf, attrib.name)
            _append_uleb128(
                buf, DW_FORM.indirect if use_dw_form_indirect else attrib.form
            )
        buf.append(0)
        buf.append(0)
        if die.children:
            for child in die.children:
                aux(child)

    for unit in units:
        aux(unit.die)
    buf.append(0)
    return buf


def _compile_debug_info(
    units, little_endian, bits, version, file_names, use_dw_form_indirect
):
    offset_size = 4  # We only emit the 32-bit format for now.
    byteorder = "little" if little_endian else "big"
    labels = {}
    references = []
    unit_references = []
    code = 1

    def aux(buf, die, depth):
        if isinstance(die, DwarfLabel):
            if die.name in labels:
                raise ValueError(f"duplicate label {die.name!r}")
            labels[die.name] = len(buf)
            return

        nonlocal code
        _append_uleb128(buf, code)
        code += 1
        for attrib in die.attribs:
            if use_dw_form_indirect:
                _append_uleb128(buf, attrib.form)
            if attrib.name == DW_AT.decl_file or attrib.name == DW_AT.call_file:
                value = file_names[attrib.value]
            else:
                value = attrib.value
            if attrib.form == DW_FORM.addr:
                buf.extend(value.to_bytes(bits // 8, byteorder))
            elif attrib.form == DW_FORM.data1:
                buf.append(value)
            elif attrib.form == DW_FORM.data2:
                buf.extend(value.to_bytes(2, byteorder))
            elif attrib.form in (DW_FORM.data4, DW_FORM.ref_sup4):
                buf.extend(value.to_bytes(4, byteorder))
            elif attrib.form in (DW_FORM.data8, DW_FORM.ref_sup8):
                buf.extend(value.to_bytes(8, byteorder))
            elif attrib.form == DW_FORM.udata:
                _append_uleb128(buf, value)
            elif attrib.form == DW_FORM.sdata:
                _append_sleb128(buf, value)
            elif attrib.form == DW_FORM.block:
                _append_uleb128(buf, len(value))
                buf.extend(value)
            elif attrib.form == DW_FORM.block1:
                buf.append(len(value))
                buf.extend(value)
            elif attrib.form in (DW_FORM.strp, DW_FORM.GNU_ref_alt):
                buf.extend(value.to_bytes(offset_size, byteorder))
            elif attrib.form == DW_FORM.string:
                buf.extend(value.encode())
                buf.append(0)
            elif attrib.form == DW_FORM.ref1:
                if isinstance(value, str):
                    unit_references.append((len(buf), 1, value))
                    buf.append(0)
                else:
                    buf.extend(value.to_bytes(1, byteorder))
            elif attrib.form == DW_FORM.ref2:
                if isinstance(value, str):
                    unit_references.append((len(buf), 2, value))
                    buf.extend(bytes(2))
                else:
                    buf.extend(value.to_bytes(2, byteorder))
            elif attrib.form == DW_FORM.ref4:
                if isinstance(value, str):
                    unit_references.append((len(buf), 4, value))
                    buf.extend(bytes(4))
                else:
                    buf.extend(value.to_bytes(4, byteorder))
            elif attrib.form == DW_FORM.ref8:
                if isinstance(value, str):
                    unit_references.append((len(buf), 8, value))
                    buf.extend(bytes(8))
                else:
                    buf.extend(value.to_bytes(8, byteorder))
            elif attrib.form == DW_FORM.ref_udata:
                if isinstance(value, str):
                    assert (
                        value in labels
                    ), "DW_FORM_ref_udata can only be used for backreferences"
                    _append_uleb128(buf, labels[value] - unit_offset)
                else:
                    _append_uleb128(buf, value)
            elif attrib.form == DW_FORM.ref_addr:
                if isinstance(value, str):
                    references.append((len(buf), offset_size, value))
                    buf.extend(bytes(offset_size))
                else:
                    buf.extend(value.to_bytes(offset_size, byteorder))
            elif attrib.form == DW_FORM.ref_sig8:
                buf.extend(value.to_bytes(8, byteorder))
            elif attrib.form == DW_FORM.sec_offset:
                buf.extend(bytes(offset_size))
            elif attrib.form == DW_FORM.flag_present:
                pass
            elif attrib.form == DW_FORM.exprloc:
                _append_uleb128(buf, len(value))
                buf.extend(value)
            else:
                assert False, attrib.form
        if die.children:
            for child in die.children:
                aux(buf, child, depth + 1)
            buf.append(0)

    debug_info = bytearray()
    debug_types = bytearray()
    for unit in units:
        unit_references.clear()
        if version == 4 and unit.type in (DW_UT.type, DW_UT.split_type):
            buf = debug_types
        else:
            buf = debug_info
        unit_offset = len(buf)
        buf.extend(b"\0\0\0\0")  # unit_length
        buf.extend(version.to_bytes(2, byteorder))  # version
        if version >= 5:
            buf.append(unit.type)  # unit_type
            buf.append(bits // 8)  # address_size
        buf.extend((0).to_bytes(4, byteorder))  # debug_abbrev_offset
        if version < 5:
            buf.append(bits // 8)  # address_size

        if version >= 5 and unit.type in (DW_UT.skeleton, DW_UT.split_compile):
            buf.extend(unit.dwo_id.to_bytes(8, byteorder))  # dwo_id
        else:
            assert unit.dwo_id is None
        if unit.type in (DW_UT.type, DW_UT.split_type):
            buf.extend(unit.type_signature.to_bytes(8, byteorder))  # type_signature
            unit_references.append((len(buf), offset_size, unit.type_offset))
            buf.extend(bytes(offset_size))  # type_offset
        else:
            assert unit.type_signature is None
            assert unit.type_offset is None

        if unit.die_label is not None:
            aux(buf, DwarfLabel(unit.die_label), 0)
        aux(buf, unit.die, 0)

        unit_length = len(buf) - unit_offset - 4
        buf[unit_offset : unit_offset + 4] = unit_length.to_bytes(4, byteorder)

        for offset, size, label in unit_references:
            die_offset = labels[label] - unit_offset
            buf[offset : offset + size] = die_offset.to_bytes(size, byteorder)

    for offset, size, label in references:
        buf[offset : offset + size] = labels[label].to_bytes(size, byteorder)

    return debug_info, debug_types, labels


def _compile_debug_line(units, little_endian, bits, version):
    byteorder = "little" if little_endian else "big"

    if not units:
        units = [DwarfUnit(DW_UT.compile, DwarfDie(DW_TAG.compile_unit, []))]

    buf = bytearray()
    for unit in units:
        lnp = unit.lnp
        if lnp is None:
            lnp = DwarfLnp(
                instructions=(),
                comp_dir="/usr/src",
                file_names=("main.c",),
            )

        unit.die.attribs.append(
            DwarfAttrib(DW_AT.stmt_list, DW_FORM.sec_offset, len(buf))
        )
        if unit.type in (DW_UT.compile, DW_UT.partial, DW_UT.skeleton):
            unit.die.attribs.append(
                DwarfAttrib(DW_AT.name, DW_FORM.string, lnp.file_names[0])
            )
            unit.die.attribs.append(
                DwarfAttrib(DW_AT.comp_dir, DW_FORM.string, lnp.comp_dir)
            )

        unit_length_start = len(buf)
        buf.extend(b"\0\0\0\0")  # unit_length
        unit_length_end = len(buf)
        buf.extend(version.to_bytes(2, byteorder))  # version
        if version >= 5:
            buf.append(bits // 8)  # address_size
            buf.append(0)  # segment_selector_size
        header_length_start = len(buf)
        buf.extend(b"\0\0\0\0")  # header_length
        header_length_end = len(buf)
        buf.append(1)  # minimum_instruction_length
        buf.append(1)  # maximum_operations_per_instruction
        buf.append(lnp.default_is_stmt)  # default_is_stmt
        buf.append(1)  # line_base
        buf.append(1)  # line_range
        buf.append(13)  # opcode_base
        buf.extend([0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 1])  # standard_opcode_lengths
        if version >= 5:
            buf.append(1)  # directory_entry_format_count
            # directory_entry_format
            _append_uleb128(buf, DW_LNCT.path)
            _append_uleb128(buf, DW_FORM.string)

        directories = {lnp.comp_dir: 0}
        file_names = {}
        file_name_entries = []

        for file_name in lnp.file_names:
            assert file_name not in file_names

            file_names[file_name] = len(file_names) + 1

            dirname, basename = os.path.split(file_name)
            if dirname:
                directory_index = directories.setdefault(dirname, len(directories))
            else:
                directory_index = 0
            file_name_entries.append((basename, directory_index))

        def collect_file_names(die):
            if isinstance(die, DwarfLabel):
                return
            for attrib in die.attribs:
                if attrib.name not in (DW_AT.decl_file, DW_AT.call_file):
                    continue

                if attrib.value in file_names:
                    continue

                file_names[attrib.value] = len(file_names) + 1

                dirname, basename = os.path.split(attrib.value)
                if dirname:
                    directory_index = directories.setdefault(dirname, len(directories))
                else:
                    directory_index = 0
                file_name_entries.append((basename, directory_index))
            for child in die.children:
                collect_file_names(child)

        collect_file_names(unit.die)

        if version >= 5:
            _append_uleb128(buf, len(directories))  # directories_count

        # directories (or include_directories in version <= 4)
        for directory, index in directories.items():
            if index > 0 or version >= 5:
                buf.extend(directory.encode("ascii"))
                buf.append(0)
        if version < 5:
            buf.append(0)

        if version >= 5:
            buf.append(2)  # file_name_entry_format_count
            # file_name_entry_format
            _append_uleb128(buf, DW_LNCT.path)
            _append_uleb128(buf, DW_FORM.string)
            _append_uleb128(buf, DW_LNCT.directory_index)
            _append_uleb128(buf, DW_FORM.udata)

        if version >= 5:
            file_name_entries.insert(0, file_name_entries[0])
            _append_uleb128(buf, len(file_names))  # file_names_count

        # file_names
        for path, directory_index in file_name_entries:
            # path
            buf.extend(path.encode("ascii"))
            buf.append(0)
            _append_uleb128(buf, directory_index)  # directory_index
            if version < 5:
                _append_uleb128(buf, 0)  # mtime
                _append_uleb128(buf, 0)  # size

        if version < 5:
            buf.append(0)

        buf[header_length_start:header_length_end] = (
            len(buf) - header_length_end
        ).to_bytes(header_length_end - header_length_start, byteorder)

        for instruction in lnp.instructions:
            opcode = instruction[0]
            buf.append(opcode)
            if opcode in {DW_LNS.copy, DW_LNE.end_sequence}:
                assert len(instruction) == 1
            elif opcode == DW_LNS.advance_pc:
                assert len(instruction) == 2
                _append_uleb128(buf, instruction[1])
            elif opcode == DW_LNS.advance_line:
                assert len(instruction) == 2
                _append_sleb128(buf, instruction[1])
            elif opcode == DW_LNS.set_file:
                assert len(instruction) == 2
                _append_uleb128(buf, instruction[1])
            elif opcode == DW_LNS.set_column:
                assert len(instruction) == 2
                _append_uleb128(buf, instruction[1])
            else:
                assert False, opcode

        buf[unit_length_start:unit_length_end] = (len(buf) - unit_length_end).to_bytes(
            unit_length_end - unit_length_start, byteorder
        )
    return buf, file_names


_UNIT_TAGS = frozenset({DW_TAG.type_unit, DW_TAG.compile_unit, DW_TAG.partial_unit})


class DwarfResult(NamedTuple):
    data: bytes
    labels: Dict[str, int]


def compile_dwarf(
    units_or_dies,
    *,
    version=4,
    lang=None,
    use_dw_form_indirect=False,
    compress=None,
    split=None,
    sections=(),
    little_endian=True,
    bits=64,
    allow_any_unit_die=False,
    **kwargs,
):
    assert compress in (None, "zlib-gnu", "zlib-gabi")
    assert split in (None, "dwo")

    if isinstance(units_or_dies, (DwarfDie, DwarfUnit)):
        units_or_dies = (units_or_dies,)
    if not units_or_dies or isinstance(units_or_dies[0], DwarfUnit):
        units = units_or_dies
    else:
        assert all(isinstance(die, (DwarfDie, DwarfLabel)) for die in units_or_dies)
        assert all(
            not isinstance(die, DwarfDie) or die.tag not in _UNIT_TAGS
            for die in units_or_dies
        )
        units = (
            DwarfUnit(DW_UT.compile, DwarfDie(DW_TAG.compile_unit, (), units_or_dies)),
        )
    assert all(isinstance(unit, DwarfUnit) for unit in units)
    assert allow_any_unit_die or all(unit.die.tag in _UNIT_TAGS for unit in units)

    unit_attribs = []
    if lang is not None:
        unit_attribs.append(DwarfAttrib(DW_AT.language, DW_FORM.data1, lang))

    units = [
        unit._replace(
            die=unit.die._replace(attribs=list(unit.die.attribs) + unit_attribs)
        )
        for unit in units
    ]

    # TODO: line number information for a split file is in the skeleton file.
    # We don't have any test cases yet that use line number information from a
    # split file, but when we do, we'll have to add a way to include the split
    # file's line number information in the skeleton file.
    if split:
        file_names = {}
    else:
        debug_line, file_names = _compile_debug_line(
            units, little_endian, bits, version
        )

    debug_info, debug_types, labels = _compile_debug_info(
        units, little_endian, bits, version, file_names, use_dw_form_indirect
    )

    def debug_section(name, data):
        assert name.startswith(".debug")
        if compress == "zlib-gnu":
            name = ".z" + name[1:]
            compressed_data = bytearray(b"ZLIB")
            compressed_data.extend(len(data).to_bytes(8, "big"))
            compressed_data.extend(zlib.compress(data))
            data = compressed_data
        if split:
            name += ".dwo"
        return ElfSection(
            name=name,
            sh_type=SHT.PROGBITS,
            data=data,
            sh_flags=SHF.COMPRESSED if compress == "zlib-gabi" else SHF(0),
        )
        return name

    dwarf_sections = [
        debug_section(
            ".debug_abbrev", _compile_debug_abbrev(units, use_dw_form_indirect)
        ),
        debug_section(".debug_info", debug_info),
        debug_section(".debug_str", b"\0"),
    ]
    if not split:
        dwarf_sections.append(debug_section(".debug_line", debug_line))
    if debug_types:
        dwarf_sections.append(debug_section(".debug_types", debug_types))

    return DwarfResult(
        data=create_elf_file(
            ET.EXEC,
            sections=[*sections, *dwarf_sections],
            little_endian=little_endian,
            bits=bits,
            **kwargs,
        ),
        labels=labels,
    )


def create_dwarf_file(*args, **kwargs):
    return compile_dwarf(*args, **kwargs).data



#!/usr/bin/env python3

from collections import deque
import concurrent.futures
import contextlib
import functools
import logging
import os
from pathlib import Path
import shlex
import shutil
import subprocess
import sys
import time
import traceback
from typing import (
    TYPE_CHECKING,
    Callable,
    Deque,
    Dict,
    List,
    Optional,
    Protocol,
    Set,
    TextIO,
    Tuple,
    Union,
)

from util import KernelVersion, nproc
from vmtest.chroot import chroot_sh_cmd
from vmtest.config import (
    ARCHITECTURES,
    HOST_ARCHITECTURE,
    KERNEL_FLAVORS,
    SUPPORTED_KERNEL_VERSIONS,
    Architecture,
    Compiler,
    Kernel,
)
from vmtest.download import Downloader
from vmtest.rootfsbuild import build_drgn_for_arch
from vmtest.vm import LostVMError, TestKmodMode, run_in_vm

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    if sys.version_info < (3, 10):
        from typing_extensions import ParamSpec
    else:
        from typing import ParamSpec  # novermin
    _P = ParamSpec("_P")


class _TestFunction(Protocol):
    def __call__(self, *, outfile: Optional[TextIO] = None) -> bool: ...


def _kdump_works(kernel: Kernel) -> bool:
    if kernel.arch.name == "aarch64":
        # kexec fails with "kexec: setup_2nd_dtb failed." on older versions.
        # See
        # http://lists.infradead.org/pipermail/kexec/2020-November/021740.html.
        return KernelVersion(kernel.release) >= KernelVersion("5.10")
    elif kernel.arch.name == "arm":
        # /proc/vmcore fails to initialize. See
        # https://lore.kernel.org/linux-debuggers/ZvxT9EmYkyFuFBH9@telecaster/T/.
        return False
    elif kernel.arch.name == "ppc64":
        # Before 6.1, sysrq-c hangs.
        return KernelVersion(kernel.release) >= KernelVersion("6.1")
    elif kernel.arch.name == "s390x":
        # Before 5.15, sysrq-c hangs.
        return KernelVersion(kernel.release) >= KernelVersion("5.15")
    elif kernel.arch.name in {"x86_64", "riscv64"}:
        return True
    else:
        assert False, kernel.arch.name


def _default_parallelism(mem_gb: float = 2, cpu: float = 1.75) -> int:
    for line in open("/proc/meminfo"):
        fields = line.split()
        if fields[0] == "MemAvailable:":
            mem_available_gb = int(fields[1]) / (1024 * 1024)
            break
    else:
        return 1

    limit_mem = mem_available_gb // mem_gb
    limit_cpu = nproc() // cpu
    return int(max(1, min(limit_mem, limit_cpu)))


class _TestRunner:
    def __init__(
        self,
        *,
        directory: Path,
        jobs: Optional[int] = None,
        use_host_rootfs: bool = True,
        skip_build: bool = False,
        pytest_kernel_args: Optional[str] = None,
        skip_kdump: bool = False,
    ) -> None:
        self._directory = directory
        if jobs is None:
            self._jobs = 1
        elif jobs == 0:
            self._jobs = _default_parallelism()
            logger.info("using default parallelism %d", self._jobs)
        else:
            self._jobs = jobs
            logger.info("using parallelism %d", self._jobs)
        self._foreground = jobs is None
        self._use_host_rootfs = use_host_rootfs
        self._skip_build = skip_build

        self._compilers_to_resolve: Dict[Architecture, None] = {}
        self._kernels_to_resolve: Dict[Tuple[Architecture, str], None] = {}
        self._drgn_builds: Dict[Architecture, None] = {}

        # + 1 for download tasks.
        self._pool = concurrent.futures.ThreadPoolExecutor(max_workers=self._jobs + 1)
        self._futures: Set["concurrent.futures.Future[Callable[[], bool]]"] = set()

        self._downloader = Downloader(directory)
        self._download_queue: Deque[Union[Compiler, Kernel]] = deque()

        self._test_queue: Deque[Tuple[str, str, _TestFunction]] = deque()
        self._tests_running: Dict[Tuple[str, str], float] = {}
        self._tests_passed: Dict[str, List[str]] = {}
        self._tests_failed: Dict[str, List[str]] = {}

        try:
            self._color = os.isatty(sys.stderr.fileno())
        except (AttributeError, OSError):
            self._color = False

        self._pytest_kernel_args = pytest_kernel_args
        self._skip_kdump = skip_kdump

    def add_kernel(self, arch: Architecture, pattern: str) -> None:
        self._compilers_to_resolve[arch] = None
        self._kernels_to_resolve[(arch, pattern)] = None
        if not self._skip_build:
            self._drgn_builds[arch] = None

    def add_local(self, arch: Architecture) -> None:
        if not self._skip_build:
            self._drgn_builds[arch] = None
        self._queue_local_test(arch)

    def _submit(
        self,
        fn: Callable["_P", Callable[[], bool]],
        *args: "_P.args",
        **kwargs: "_P.kwargs",
    ) -> None:
        self._futures.add(self._pool.submit(fn, *args, **kwargs))

    def run(self) -> bool:
        try:
            self._submit(self._resolve_downloads)

            self._submit_next_drgn_build()

            self._print_progress()
            while self._futures:
                done, self._futures = concurrent.futures.wait(
                    self._futures,
                    timeout=None if self._foreground else 1,
                    return_when=concurrent.futures.FIRST_COMPLETED,
                )
                update_progress = not self._foreground
                for future in done:
                    callback = future.result()
                    update_progress |= callback()
                if update_progress:
                    self._print_progress()
        except Exception:
            traceback.print_exc()
            return False
        finally:
            for future in self._futures:
                future.cancel()
            self._pool.shutdown()
        return not self._tests_failed

    def _green(self, s: str) -> str:
        if self._color:
            return "\033[32m" + s + "\033[m"
        else:
            return s

    def _red(self, s: str) -> str:
        if self._color:
            return "\033[31m" + s + "\033[m"
        else:
            return s

    def _yellow(self, s: str) -> str:
        if self._color:
            return "\033[33m" + s + "\033[m"
        else:
            return s

    def _cyan(self, s: str) -> str:
        if self._color:
            return "\033[36m" + s + "\033[m"
        else:
            return s

    def _print_progress(self) -> None:
        parts = []
        if self._foreground:
            endl = "\n"
        else:
            # To minimize flicker, we overwrite the output instead of clearing.
            parts.append("\033[H")  # Move cursor to top left corner.
            endl = "\033[K\n"  # Clear to the end of line on each newline.
            if self._compilers_to_resolve or self._kernels_to_resolve:
                parts.append(self._cyan("Queueing downloads..."))
                parts.append(endl)
            elif self._download_queue:
                num_compilers = sum(
                    isinstance(download, Compiler) for download in self._download_queue
                )
                num_kernels = len(self._download_queue) - num_compilers

                downloading_parts = []
                if num_compilers == 1:
                    downloading_parts.append("1 compiler")
                elif num_compilers > 1:
                    downloading_parts.append(f"{num_compilers} compilers")
                if num_kernels == 1:
                    downloading_parts.append("1 kernel")
                elif num_kernels > 1:
                    downloading_parts.append(f"{num_kernels} kernels")

                parts.append(
                    self._cyan(f"Downloading {' and '.join(downloading_parts)}...")
                )
                parts.append(endl)

            if self._test_queue:
                parts.append(self._cyan(f"{len(self._test_queue)} tests waiting..."))
                parts.append(endl)

            if self._drgn_builds:
                parts.append(self._yellow("Building: "))
                parts.append(", ".join([arch.name for arch in self._drgn_builds]))
                parts.append(endl)

            now = time.monotonic()
            first = True
            for (category_name, test_name), start_time in reversed(
                self._tests_running.items()
            ):
                if first:
                    parts.append(self._yellow("Running: "))
                    first = False
                else:
                    parts.append("         ")
                parts.append(f"{category_name}: {test_name} ({int(now - start_time)}s)")
                parts.append(endl)

        for title, results, color in (
            ("Passed", self._tests_passed, self._green),
            ("Failed", self._tests_failed, self._red),
        ):
            first = True
            for category_name, test_names in sorted(results.items()):
                if first:
                    parts.append(color(title + ":"))
                    parts.append(" ")
                    first = False
                else:
                    parts.append(" " * (len(title) + 2))
                parts.append(f"{category_name}: {', '.join(test_names)}")
                parts.append(endl)

        if not self._foreground:
            parts.append("\033[J")  # Clear the rest of the screen.
        sys.stderr.write("".join(parts))

    def _submit_next_drgn_build(self) -> None:
        if self._drgn_builds:
            self._submit(self._build_drgn, next(iter(self._drgn_builds)))
        else:
            self._submit_tests()

    def _rootfs(self, arch: Architecture) -> Path:
        if self._use_host_rootfs and arch is HOST_ARCHITECTURE:
            return Path("/")
        else:
            return self._directory / arch.name / "rootfs"

    def _build_drgn(self, arch: Architecture) -> Callable[[], bool]:
        with contextlib.ExitStack() as exit_stack:
            if self._foreground:
                outfile = None
            else:
                outfile = exit_stack.enter_context(
                    (self._directory / "log" / f"{arch.name}-build.log").open("w")
                )
            if self._use_host_rootfs and arch is HOST_ARCHITECTURE:
                subprocess.check_call(
                    [sys.executable, "setup.py", "build_ext", "-i"],
                    stdout=outfile,
                    stderr=outfile,
                )
            else:
                build_drgn_for_arch(arch, self._directory, outfile=outfile)
        return functools.partial(self._drgn_build_done, arch)

    def _drgn_build_done(self, arch: Architecture) -> bool:
        del self._drgn_builds[arch]
        self._submit_next_drgn_build()
        return not self._foreground

    def _resolve_downloads(self) -> Callable[[], bool]:
        for target in self._compilers_to_resolve:
            compiler = self._downloader.resolve_compiler(target)
            self._download_queue.append(compiler)

        for arch, pattern in self._kernels_to_resolve:
            kernel = self._downloader.resolve_kernel(arch, pattern)
            self._download_queue.append(kernel)

        return self._resolved_downloads

    def _resolved_downloads(self) -> bool:
        self._compilers_to_resolve.clear()
        self._kernels_to_resolve.clear()
        return self._submit_next_download()

    def _submit_next_download(self) -> bool:
        if self._download_queue:
            self._submit(self._download, self._download_queue[0])
        return not self._foreground

    def _download(self, download: Union[Compiler, Kernel]) -> Callable[[], bool]:
        if isinstance(download, Compiler):
            self._downloader.download_compiler(download)
        else:
            self._downloader.download_kernel(download)
        return functools.partial(self._download_done, download)

    def _download_done(self, download: Union[Compiler, Kernel]) -> bool:
        popped = self._download_queue.popleft()
        assert popped is download
        self._submit_next_download()
        if isinstance(download, Kernel):
            self._queue_kernel_test(download)
        return not self._foreground

    def _queue_local_test(self, arch: Architecture) -> None:
        self._queue_test(arch.name, "local", functools.partial(self._test_local, arch))

    def _queue_kernel_test(self, kernel: Kernel) -> None:
        self._queue_test(
            kernel.arch.name,
            kernel.release,
            functools.partial(self._test_kernel, kernel),
        )

    def _queue_test(
        self, category_name: str, test_name: str, fn: _TestFunction
    ) -> None:
        self._test_queue.append((category_name, test_name, fn))
        logger.info("%s %s test queued", category_name, test_name)
        if not self._drgn_builds:
            self._submit_tests()

    def _submit_tests(self) -> None:
        assert not self._drgn_builds
        while self._test_queue and len(self._tests_running) < self._jobs:
            category_name, test_name, fn = self._test_queue.popleft()
            self._tests_running[(category_name, test_name)] = time.monotonic()
            logger.info("%s %s test started", category_name, test_name)
            self._submit(self._test_wrapper, category_name, test_name, fn)

    def _test_wrapper(
        self, category_name: str, test_name: str, fn: _TestFunction
    ) -> Callable[[], bool]:
        with contextlib.ExitStack() as exit_stack:
            if self._foreground:
                outfile = None
            else:
                outfile = exit_stack.enter_context(
                    (self._directory / "log" / f"{category_name}-{test_name}.log").open(
                        "w"
                    )
                )
            success = fn(outfile=outfile)
        return functools.partial(self._test_done, category_name, test_name, success)

    def _test_done(self, category_name: str, test_name: str, success: bool) -> bool:
        start_time = self._tests_running.pop((category_name, test_name))
        logger.info(
            "%s %s test %s (%ds)",
            category_name,
            test_name,
            "passed" if success else "failed",
            time.monotonic() - start_time,
        )
        (self._tests_passed if success else self._tests_failed).setdefault(
            category_name, []
        ).append(test_name)
        self._submit_tests()
        return True

    def _test_local(
        self, arch: Architecture, *, outfile: Optional[TextIO] = None
    ) -> bool:
        rootfs = self._rootfs(arch)
        if rootfs == Path("/"):
            args = [
                sys.executable,
                "-m",
                "pytest",
                "-v",
                "--ignore=tests/linux_kernel",
            ]
        else:
            args = [
                "unshare",
                "--map-root-user",
                "--map-auto",
                "--fork",
                "--pid",
                "--mount-proc=" + str(rootfs / "proc"),
                "sh",
                "-c",
                f"""\
set -e

mount --bind . "$1/mnt"
{chroot_sh_cmd('"$1"')} 'cd /mnt && pytest -v --ignore=tests/linux_kernel'
""",
                "sh",
                str(rootfs),
            ]
        return subprocess.call(args, stdout=outfile, stderr=outfile) == 0

    def _test_kernel(self, kernel: Kernel, *, outfile: Optional[TextIO] = None) -> bool:
        rootfs = self._rootfs(kernel.arch)
        if rootfs == Path("/"):
            python_executable = sys.executable
        else:
            python_executable = "/usr/bin/python3"

        if _kdump_works(kernel) and not self._skip_kdump:
            kdump_command = """\
    "$PYTHON" -Bm vmtest.enter_kdump
    # We should crash and not reach this.
    exit 1
"""
        else:
            kdump_command = ""

        if self._pytest_kernel_args is not None:
            pytest_args = self._pytest_kernel_args
        elif kernel.arch is HOST_ARCHITECTURE:
            pytest_args = "tests/linux_kernel --ignore tests/linux_kernel/vmcore"
        else:
            pytest_args = (
                "tests/linux_kernel --ignore tests/linux_kernel/vmcore -m 'not slow'"
            )

        test_command = rf"""
set -e

export PYTHON={shlex.quote(python_executable)}
export DRGN_RUN_LINUX_KERNEL_TESTS=1
if [ -e /proc/vmcore ]; then
    "$PYTHON" -Bm pytest -v tests/linux_kernel/vmcore
else
    insmod "$DRGN_TEST_KMOD"
    "$PYTHON" -Bm pytest -v {pytest_args}
{kdump_command}
fi
"""

        # Silence as much boot text as possible if we're specifying tests to run
        # in the foreground
        extra_kernel_cmdline = []
        if self._pytest_kernel_args and self._foreground:
            extra_kernel_cmdline.append("quiet")

        try:
            status = run_in_vm(
                test_command,
                kernel,
                rootfs,
                self._directory,
                test_kmod=TestKmodMode.BUILD,
                outfile=outfile,
                extra_kernel_cmdline=extra_kernel_cmdline,
            )
            return status == 0
        except (
            LostVMError,
            subprocess.CalledProcessError,  # For kmod build errors.
        ) as e:
            print(e, file=sys.stderr if outfile is None else outfile)
            return False


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="test drgn in a virtual machine",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-d",
        "--directory",
        metavar="DIR",
        type=Path,
        default="build/vmtest",
        help="directory for vmtest artifacts",
    )
    parser.add_argument(
        "-a",
        "--architecture",
        dest="architectures",
        action="append",
        choices=["all", "foreign", *sorted(ARCHITECTURES)],
        default=argparse.SUPPRESS,
        required=HOST_ARCHITECTURE is None,
        help="architecture to test, "
        '"all" to test all supported architectures, '
        'or "foreign" to test all supported architectures other than the host architecture; '
        "may be given multiple times"
        + (
            "" if HOST_ARCHITECTURE is None else f" (default: {HOST_ARCHITECTURE.name})"
        ),
    )
    parser.add_argument(
        "-k",
        "--kernel",
        metavar="PATTERN|{all," + ",".join(KERNEL_FLAVORS) + "}",
        dest="kernels",
        action="append",
        default=argparse.SUPPRESS,
        help="kernel to test, "
        '"all" to test all supported kernels, '
        "or flavor name to test all supported kernels of a specific flavor; "
        "may be given multiple times (default: none)",
    )
    parser.add_argument(
        "-l",
        "--local",
        action="store_true",
        help="run local tests",
    )
    parser.add_argument(
        "-j",
        "--jobs",
        type=int,
        nargs="?",
        default=argparse.SUPPRESS,
        help="number of tests to run in parallel (default: 1). "
        "If the argument is omitted or 0, "
        "an appropriate number is chosen automatically",
    )
    parser.add_argument(
        "--use-host-rootfs",
        choices=["never", "auto"],
        default="auto",
        help='if "never", use $directory/$arch/rootfs even for host architecture; '
        'if "auto", use / for host architecture',
    )
    parser.add_argument(
        "--skip-build",
        action="store_true",
        help="don't rebuild drgn even if it's out of date",
    )
    parser.add_argument(
        "--pytest-args",
        help="a string of args passed to pytest for kernel tests",
    )
    parser.add_argument(
        "--skip-kdump",
        action="store_true",
        help="skip kdump and vmcore tests (useful with --pytest-args)",
    )
    args = parser.parse_args()

    if not hasattr(args, "kernels") and not args.local:
        parser.error("at least one of -k/--kernel or -l/--local is required")

    if hasattr(args, "jobs"):
        if args.jobs is None:
            args.jobs = 0

        log_directory = args.directory / "log"
        log_old_directory = args.directory / "log.old"

        try:
            shutil.rmtree(log_old_directory)
        except FileNotFoundError:
            pass
        try:
            log_directory.rename(log_old_directory)
        except FileNotFoundError:
            pass
        log_directory.mkdir(parents=True)

        main_log_path = log_directory / "main.log"
    else:
        args.jobs = None
        main_log_path = None
    logging.basicConfig(
        format="%(asctime)s:%(levelname)s:%(name)s:%(message)s",
        level=logging.INFO,
        filename=main_log_path,
    )

    architectures: Dict[Architecture, None] = {}
    if hasattr(args, "architectures"):
        for name in args.architectures:
            if name == "all":
                for arch in ARCHITECTURES.values():
                    architectures[arch] = None
            elif name == "foreign":
                for arch in ARCHITECTURES.values():
                    if arch is not HOST_ARCHITECTURE:
                        architectures[arch] = None
            else:
                architectures[ARCHITECTURES[name]] = None
    else:
        assert HOST_ARCHITECTURE is not None
        architectures = {HOST_ARCHITECTURE: None}

    runner = _TestRunner(
        directory=args.directory,
        jobs=args.jobs,
        use_host_rootfs=args.use_host_rootfs == "auto",
        skip_build=args.skip_build,
        pytest_kernel_args=args.pytest_args,
        skip_kdump=args.skip_kdump,
    )

    if hasattr(args, "kernels"):
        for pattern in args.kernels:
            if pattern == "all":
                for version in SUPPORTED_KERNEL_VERSIONS:
                    for arch in architectures:
                        if arch.kernel_version_supported(version):
                            for flavor in KERNEL_FLAVORS.values():
                                runner.add_kernel(arch, version + ".*" + flavor.name)
            elif pattern in KERNEL_FLAVORS:
                flavor = KERNEL_FLAVORS[pattern]
                for version in SUPPORTED_KERNEL_VERSIONS:
                    for arch in architectures:
                        if arch.kernel_version_supported(version):
                            runner.add_kernel(arch, version + ".*" + flavor.name)
            else:
                for arch in architectures:
                    runner.add_kernel(arch, pattern)

    if args.local:
        for arch in architectures:
            runner.add_local(arch)

    success = runner.run()
    sys.exit(0 if success else 1)



# Copyright (c) 2026 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn.helpers.linux.pid import find_task
from drgn.helpers.linux.stack import StackKind, kernel_stack_trace
from tests.linux_kernel import (
    LinuxKernelTestCase,
    skip_unless_have_stack_tracing,
    skip_unless_have_test_kmod,
)


class TestStack(LinuxKernelTestCase):
    @skip_unless_have_test_kmod
    @skip_unless_have_stack_tracing
    def test_kernel_stack_trace(self):
        trace = kernel_stack_trace(self.prog["drgn_test_kthread"])
        self.assertEqual(trace.prog, self.prog)
        self.assertEqual(trace.segments[0].kind, StackKind.TASK)

    @skip_unless_have_test_kmod
    @skip_unless_have_stack_tracing
    def test_kernel_stack_trace_by_pid_object(self):
        trace = kernel_stack_trace(self.prog["drgn_test_kthread"].pid)
        self.assertEqual(trace.prog, self.prog)
        self.assertEqual(trace.segments[0].kind, StackKind.TASK)

    @skip_unless_have_test_kmod
    @skip_unless_have_stack_tracing
    def test_kernel_stack_trace_by_pid_int(self):
        trace = kernel_stack_trace(self.prog, int(self.prog["drgn_test_kthread"].pid))
        self.assertEqual(trace.prog, self.prog)
        self.assertEqual(trace.segments[0].kind, StackKind.TASK)

    @skip_unless_have_stack_tracing
    def test_kernel_stack_trace_user(self):
        task = find_task(self.prog, 1)
        try:
            trace = kernel_stack_trace(task)
        except ValueError as e:
            if "cannot unwind stack of running task" in str(e):
                self.skipTest("init task is currently on CPU")
            else:
                raise

        self.assertEqual(trace.prog, self.prog)
        self.assertEqual(trace.segments[0].kind, StackKind.TASK)

        # Not all architectures can unwind into the userspace frame. But if so,
        # we should categorize it correctly.
        if len(trace.segments) > 1:
            self.assertEqual(trace.segments[1].kind, StackKind.USER)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
from contextlib import contextmanager
import fnmatch
import functools
import glob
import logging
import os
from pathlib import Path
import queue
import re
import shutil
import subprocess
import tempfile
import threading
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Union,
)
import urllib.request

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from util import KernelVersion
from vmtest.config import (
    ARCHITECTURES,
    HOST_ARCHITECTURE,
    KERNEL_ORG_COMPILER_VERSION,
    Architecture,
    Compiler,
    Kernel,
    compiler_name,
    compiler_url,
    local_kernel,
)
from vmtest.githubapi import GitHubApi

logger = logging.getLogger(__name__)

VMTEST_GITHUB_RELEASE = ("osandov", "drgn", "vmtest-assets")


GitHubAsset = Dict[str, Any]


def available_kernel_releases(
    github_release: Dict[str, Any]
) -> Dict[str, Dict[str, GitHubAsset]]:
    releases: Dict[str, Dict[str, GitHubAsset]] = {}
    pattern = re.compile(r"kernel-(?P<release>.*)\.(?P<arch>\w+)\.tar\.zst")
    for asset in github_release["assets"]:
        match = pattern.fullmatch(asset["name"])
        if match:
            try:
                arch_releases = releases[match.group("arch")]
            except KeyError:
                arch_releases = releases[match.group("arch")] = {}
            arch_releases[match.group("release")] = asset
    return releases


class DownloadKernel(NamedTuple):
    arch: Architecture
    pattern: str


class DownloadCompiler(NamedTuple):
    target: Architecture


Download = Union[DownloadKernel, DownloadCompiler]
Downloaded = Union[Kernel, Compiler]


class DownloadNotFoundError(Exception):
    pass


_KERNEL_ORG_COMPILER_HOST_NAME = (
    None
    if HOST_ARCHITECTURE is None
    else HOST_ARCHITECTURE.kernel_org_compiler_host_name
)


def downloaded_compiler(download_dir: Path, target: Architecture) -> Compiler:
    if _KERNEL_ORG_COMPILER_HOST_NAME is None:
        raise DownloadNotFoundError(
            f"kernel.org compilers are not available for {NORMALIZED_MACHINE_NAME} hosts"
        )
    return Compiler(
        target=target,
        bin=download_dir
        / compiler_name(
            _KERNEL_ORG_COMPILER_HOST_NAME, target.kernel_org_compiler_target_name
        )
        / "bin",
        prefix=target.kernel_org_compiler_target_name + "-",
    )


class Downloader:
    def __init__(self, directory: Path) -> None:
        self._directory = directory
        self._gh = GitHubApi(os.getenv("GITHUB_TOKEN"))
        self._cached_kernel_releases: Optional[Dict[str, Dict[str, GitHubAsset]]] = None

    def _available_kernel_releases(self) -> Dict[str, Dict[str, GitHubAsset]]:
        if self._cached_kernel_releases is None:
            logger.info("getting available kernel releases")
            self._directory.mkdir(parents=True, exist_ok=True)
            self._cached_kernel_releases = available_kernel_releases(
                self._gh.get_release_by_tag(
                    *VMTEST_GITHUB_RELEASE,
                    cache=self._directory / "github_release.json",
                ),
            )
        return self._cached_kernel_releases

    def resolve_kernel(self, arch: Architecture, pattern: str) -> Kernel:
        if pattern.startswith(".") or pattern.startswith("/"):
            return local_kernel(arch, Path(pattern))
        elif pattern == glob.escape(pattern):
            release = pattern
        else:
            try:
                release = max(
                    (
                        available
                        for available in self._available_kernel_releases()[arch.name]
                        if fnmatch.fnmatch(available, pattern)
                    ),
                    key=KernelVersion,
                )
            except ValueError:
                raise DownloadNotFoundError(
                    f"no available kernel release matches {pattern!r} on {arch.name}"
                )
            else:
                logger.info(
                    "kernel release pattern %s matches %s on %s",
                    pattern,
                    release,
                    arch.name,
                )
        kernel_dir = self._directory / arch.name / ("kernel-" + release)
        if (
            not kernel_dir.exists()
            and release not in self._available_kernel_releases()[arch.name]
        ):
            raise DownloadNotFoundError(
                f"kernel release {release} not found on {arch.name}"
            )
        return Kernel(arch, release, kernel_dir)

    def download_kernel(self, kernel: Kernel) -> Kernel:
        if kernel.path.exists():
            # As a policy, vmtest assets will never be updated with the same
            # name. Therefore, if the kernel was previously downloaded, we
            # don't need to download it again.
            logger.info(
                "kernel release %s for %s already downloaded to %s",
                kernel.release,
                kernel.arch.name,
                kernel.path,
            )
            return kernel

        url = self._available_kernel_releases()[kernel.arch.name][kernel.release]["url"]
        logger.info(
            "downloading kernel release %s for %s to %s from %s",
            kernel.release,
            kernel.arch.name,
            kernel.path,
            url,
        )
        kernel.path.parent.mkdir(parents=True, exist_ok=True)
        tmp_dir = Path(tempfile.mkdtemp(dir=kernel.path.parent))
        try:
            # Don't assume that the available version of tar has zstd support or
            # the non-standard -I/--use-compress-program option.
            with subprocess.Popen(
                ["zstd", "-d", "-", "--stdout"],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
            ) as zstd_proc, subprocess.Popen(
                ["tar", "-C", str(tmp_dir), "-x"],
                stdin=zstd_proc.stdout,
            ) as tar_proc:
                assert zstd_proc.stdin is not None
                try:
                    with self._gh.download(url) as resp:
                        shutil.copyfileobj(resp, zstd_proc.stdin)
                finally:
                    zstd_proc.stdin.close()
            if zstd_proc.returncode != 0:
                raise subprocess.CalledProcessError(
                    zstd_proc.returncode, zstd_proc.args
                )
            if tar_proc.returncode != 0:
                raise subprocess.CalledProcessError(tar_proc.returncode, tar_proc.args)
        except BaseException:
            shutil.rmtree(tmp_dir, ignore_errors=True)
            raise
        else:
            tmp_dir.rename(kernel.path)
        return kernel

    def resolve_compiler(self, target: Architecture) -> Compiler:
        return downloaded_compiler(self._directory, target)

    def download_compiler(self, compiler: Compiler) -> Compiler:
        dir = compiler.bin.parent
        if dir.exists():
            logger.info(
                "compiler for %s already downloaded to %s", compiler.target.name, dir
            )
        else:
            assert _KERNEL_ORG_COMPILER_HOST_NAME is not None
            url = compiler_url(
                _KERNEL_ORG_COMPILER_HOST_NAME,
                compiler.target.kernel_org_compiler_target_name,
            )
            logger.info(
                "downloading compiler for %s from %s to %s",
                compiler.target.name,
                url,
                dir,
            )
            dir.parent.mkdir(parents=True, exist_ok=True)
            with tempfile.TemporaryDirectory(dir=dir.parent) as tmp_name:
                tmp_dir = Path(tmp_name)
                with subprocess.Popen(
                    ["xz", "--decompress"],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                ) as xz_proc, subprocess.Popen(
                    ["tar", "-C", str(tmp_dir), "-x"],
                    stdin=xz_proc.stdout,
                ) as tar_proc:
                    assert xz_proc.stdin is not None
                    try:
                        with urllib.request.urlopen(url) as resp:
                            shutil.copyfileobj(resp, xz_proc.stdin)
                    finally:
                        xz_proc.stdin.close()
                if xz_proc.returncode != 0:
                    raise subprocess.CalledProcessError(
                        xz_proc.returncode, xz_proc.args
                    )
                if tar_proc.returncode != 0:
                    raise subprocess.CalledProcessError(
                        tar_proc.returncode, tar_proc.args
                    )
                archive_subdir = Path(
                    f"gcc-{KERNEL_ORG_COMPILER_VERSION}-nolibc/{compiler.target.kernel_org_compiler_target_name}"
                )
                archive_bin_subdir = archive_subdir / "bin"
                if not (tmp_dir / archive_bin_subdir).exists():
                    raise FileNotFoundError(
                        f"downloaded archive does not contain {archive_bin_subdir}"
                    )
                (tmp_dir / archive_subdir).rename(dir)
        return compiler


def _download_thread(
    download_dir: Path,
    downloads: Iterable[Download],
    q: "queue.Queue[Union[Downloaded, Exception]]",
) -> None:
    try:
        downloader = Downloader(download_dir)
        download_calls: List[Callable[[], Downloaded]] = []

        for download in downloads:
            if isinstance(download, DownloadKernel):
                kernel = downloader.resolve_kernel(download.arch, download.pattern)
                download_calls.append(
                    functools.partial(downloader.download_kernel, kernel)
                )
            elif isinstance(download, DownloadCompiler):
                compiler = downloader.resolve_compiler(download.target)
                download_calls.append(
                    functools.partial(downloader.download_compiler, compiler)
                )
            else:
                assert False

        for call in download_calls:
            q.put(call())
    except Exception as e:
        q.put(e)


@contextmanager
def download_in_thread(
    download_dir: Path, downloads: Iterable[Download], max_pending_kernels: int = 0
) -> Generator[Iterator[Downloaded], None, None]:
    q: "queue.Queue[Union[Downloaded, Exception]]" = queue.Queue(
        maxsize=max_pending_kernels
    )

    def aux() -> Iterator[Downloaded]:
        while True:
            obj = q.get()
            if isinstance(obj, StopIteration):
                break
            elif isinstance(obj, Exception):
                raise obj
            yield obj

    thread = None
    try:
        thread = threading.Thread(
            target=_download_thread,
            args=(download_dir, downloads, q),
            daemon=True,
        )
        thread.start()
        yield aux()
    finally:
        if thread:
            thread.join()


ARCH_ARGPARSE_METAVAR = f"{{{','.join(ARCHITECTURES)}}}"
if HOST_ARCHITECTURE is None:
    DOWNLOAD_KERNEL_ARGPARSE_METAVAR = f"{ARCH_ARGPARSE_METAVAR}:PATTERN"
    DEFAULT_ARCH_ARGPARSE_HELP = ""
else:
    DOWNLOAD_KERNEL_ARGPARSE_METAVAR = f"[{ARCH_ARGPARSE_METAVAR}:]PATTERN"
    DEFAULT_ARCH_ARGPARSE_HELP = f" (default: {HOST_ARCHITECTURE.name})"


def architecture_argparse_type(arg: str) -> Architecture:
    try:
        return ARCHITECTURES[arg]
    except KeyError:
        raise argparse.ArgumentTypeError(
            f"architecture must be one of ({', '.join(ARCHITECTURES)})"
        ) from None


def download_kernel_argparse_type(arg: str) -> DownloadKernel:
    arch_name, sep, pattern = arg.rpartition(":")
    if sep:
        arch = architecture_argparse_type(arch_name)
    else:
        if HOST_ARCHITECTURE is None:
            raise argparse.ArgumentTypeError("architecture is required")
        arch = HOST_ARCHITECTURE
    return DownloadKernel(arch, pattern)


def main() -> None:
    logging.basicConfig(
        format="%(asctime)s:%(levelname)s:%(name)s:%(message)s", level=logging.INFO
    )

    parser = argparse.ArgumentParser(
        description="Download drgn vmtest assets",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-k",
        "--kernel",
        metavar=DOWNLOAD_KERNEL_ARGPARSE_METAVAR,
        dest="downloads",
        action="append",
        type=download_kernel_argparse_type,
        default=argparse.SUPPRESS,
        help=f"download latest kernel for given architecture{DEFAULT_ARCH_ARGPARSE_HELP} matching glob pattern; may be given multiple times",
    )
    parser.add_argument(
        "-c",
        "--compiler",
        metavar=ARCH_ARGPARSE_METAVAR,
        dest="downloads",
        action="append",
        nargs=1 if HOST_ARCHITECTURE is None else "?",
        type=lambda arg: DownloadCompiler(architecture_argparse_type(arg)),
        default=argparse.SUPPRESS,
        help=f"download compiler for given architecture{DEFAULT_ARCH_ARGPARSE_HELP}; may be given multiple times",
    )
    parser.add_argument(
        "-d",
        "--download-directory",
        metavar="DIR",
        type=Path,
        default="build/vmtest",
        help="directory to download assets to",
    )
    args = parser.parse_args()
    if not hasattr(args, "downloads"):
        args.downloads = []
    # --compiler with no argument is appended as None. Fix it.
    for i, download_arg in enumerate(args.downloads):
        if download_arg is None:
            assert HOST_ARCHITECTURE is not None
            args.downloads[i] = DownloadCompiler(HOST_ARCHITECTURE)

    downloader = Downloader(args.download_directory)
    for download in args.downloads:
        if isinstance(download, DownloadKernel):
            kernel = downloader.download_kernel(
                downloader.resolve_kernel(download.arch, download.pattern)
            )
            print(
                f"kernel: arch={kernel.arch.name} release={kernel.release} path={kernel.path}"
            )
        elif isinstance(download, DownloadCompiler):
            compiler = downloader.download_compiler(
                downloader.resolve_compiler(download.target)
            )
            print(
                f"compiler: target={compiler.target.name} bin={compiler.bin} prefix={compiler.prefix}"
            )
        else:
            assert False


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn.helpers.linux.cpumask import for_each_possible_cpu
from drgn.helpers.linux.percpu import (
    per_cpu,
    per_cpu_ptr,
    percpu_counter_sum,
    percpu_counter_sum_positive,
)
from tests.linux_kernel import LinuxKernelTestCase, prng32, skip_unless_have_test_kmod


class TestPerCpu(LinuxKernelTestCase):
    def test_per_cpu(self):
        for cpu in for_each_possible_cpu(self.prog):
            rq = per_cpu(self.prog["runqueues"], cpu)
            try:
                rq_cpu = rq.cpu
            except AttributeError:
                # Before Linux kernel commit cac5cefbade9 ("sched/smp: Make SMP
                # unconditional") (in v6.17), struct rq::cpu only exists if
                # CONFIG_SMP=y, so check this instead.
                self.assertEqual(rq.idle, self.prog["init_task"].address_of_())
            else:
                self.assertEqual(rq_cpu, cpu)

    @skip_unless_have_test_kmod
    def test_per_cpu_module_static(self):
        for cpu, expected in zip(for_each_possible_cpu(self.prog), prng32("PCPU")):
            self.assertEqual(
                per_cpu(self.prog["drgn_test_percpu_static"], cpu), expected
            )

    @skip_unless_have_test_kmod
    def test_per_cpu_module_dynamic(self):
        for cpu, expected in zip(for_each_possible_cpu(self.prog), prng32("pcpu")):
            self.assertEqual(
                per_cpu_ptr(self.prog["drgn_test_percpu_dynamic"], cpu)[0], expected
            )


@skip_unless_have_test_kmod
class TestPercpuCounter(LinuxKernelTestCase):
    def test_percpu_counter_sum(self):
        self.assertEqual(percpu_counter_sum(self.prog["drgn_test_percpu_counter"]), 13)
        self.assertEqual(
            percpu_counter_sum(self.prog["drgn_test_percpu_counter_negative"]), -66
        )

    def test_percpu_counter_sum_positive(self):
        self.assertEqual(
            percpu_counter_sum_positive(self.prog["drgn_test_percpu_counter"]), 13
        )
        self.assertEqual(
            percpu_counter_sum_positive(self.prog["drgn_test_percpu_counter_negative"]),
            0,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later


import http.server
import os
import os.path
from pathlib import Path
import re
import shutil
import socket
import socketserver
import tempfile
import threading

from drgn import ModuleFileStatus, Program, SupplementaryFileKind
from tests import TestCase, modifyenv
from tests.test_debug_info import NamedTemporaryElfFile


class _DebuginfodHTTPHandler(http.server.BaseHTTPRequestHandler):
    def do_GET(self):
        match = re.fullmatch(
            r"/buildid/((?:[0-9a-fA-F][0-9a-fA-F])+)/(executable|debuginfo)", self.path
        )
        if not match:
            self.send_error(http.HTTPStatus.BAD_REQUEST)
            return

        build_id = bytes.fromhex(match.group(1))
        type = match.group(2)

        try:
            file_path = self.server.build_ids[build_id][type]
        except KeyError:
            self.send_error(http.HTTPStatus.NOT_FOUND)
            return

        try:
            f = open(file_path, "rb")
        except OSError:
            self.send_error(http.HTTPStatus.INTERNAL_SERVER_ERROR)
            return

        with f:
            self.send_response(http.HTTPStatus.OK)
            st = os.fstat(f.fileno())
            self.send_header("Content-Type", "application/octet-stream")
            self.send_header("Content-Length", str(st.st_size))
            self.send_header("X-Debuginfod-Size", str(st.st_size))
            self.send_header("Last-Modified", self.date_time_string(st.st_mtime))
            self.end_headers()
            shutil.copyfileobj(f, self.wfile)


class TestDebuginfodDebugInfoFinder(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.server = socketserver.TCPServer(("localhost", 0), _DebuginfodHTTPHandler)
        cls.server.build_ids = {}
        cls.server_thread = threading.Thread(
            target=cls.server.serve_forever, daemon=True
        )
        cls.server_thread.start()

    @classmethod
    def tearDownClass(cls):
        # By default, serve_forever() only checks if it should shut down every
        # 0.5 seconds. Shutting down the socket makes it check immediately.
        cls.server.socket.shutdown(socket.SHUT_RD)
        cls.server.shutdown()
        cls.server_thread.join()
        cls.server.server_close()

    def setUp(self):
        self.prog = Program()
        try:
            self.prog.set_enabled_debug_info_finders(["debuginfod"])
        except ValueError:
            self.skipTest("no debuginfod support")

        self.server.build_ids.clear()
        self.cache_dir = Path(
            self.enterContext(tempfile.TemporaryDirectory(prefix="debuginfod-cache-"))
        )
        self.enterContext(
            modifyenv(
                {
                    "DEBUGINFOD_URLS": "http://{}:{}/".format(
                        *self.server.server_address
                    ),
                    "DEBUGINFOD_CACHE_PATH": str(self.cache_dir),
                }
            )
        )

    def test_no_build_id(self):
        module = self.prog.extra_module("foo", create=True)
        self.prog.load_module_debug_info(module)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)

    def test_separate(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        with NamedTemporaryElfFile(
            loadable=True, debug=False, build_id=build_id
        ) as loadable_file, NamedTemporaryElfFile(
            loadable=False, debug=True, build_id=build_id
        ) as debug_file:
            self.server.build_ids[build_id] = {
                "executable": loadable_file.name,
                "debuginfo": debug_file.name,
            }

            module = self.prog.extra_module("foo", create=True)
            module.build_id = build_id
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.loaded_file_path,
                str(self.cache_dir / build_id.hex() / "executable"),
            )
            self.assertEqual(
                module.debug_file_path,
                str(self.cache_dir / build_id.hex() / "debuginfo"),
            )

    def test_no_servers(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        with NamedTemporaryElfFile(
            loadable=True, debug=False, build_id=build_id
        ) as loadable_file, NamedTemporaryElfFile(
            loadable=False, debug=True, build_id=build_id
        ) as debug_file, modifyenv(
            {"DEBUGINFOD_URLS": None}
        ):
            self.server.build_ids[build_id] = {
                "executable": loadable_file.name,
                "debuginfo": debug_file.name,
            }

            module = self.prog.extra_module("foo", create=True)
            module.build_id = build_id
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)

    def test_cache_hit(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"

        with NamedTemporaryElfFile(
            loadable=False, debug=True, build_id=build_id
        ) as debug_file:
            self.server.build_ids[build_id] = {"debuginfo": debug_file.name}

            for i in range(2):
                module = self.prog.extra_module("foo", i, create=True)
                module.build_id = build_id
                self.prog.load_module_debug_info(module)
                self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
                self.assertEqual(
                    module.debug_file_path,
                    str(self.cache_dir / build_id.hex() / "debuginfo"),
                )

    def test_gnu_debugaltlink(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with NamedTemporaryElfFile(
            loadable=True, debug=False, build_id=build_id
        ) as loadable_file, NamedTemporaryElfFile(
            loadable=False,
            debug=True,
            build_id=build_id,
            gnu_debugaltlink=("alt.debug", alt_build_id),
        ) as debug_file, NamedTemporaryElfFile(
            loadable=False, debug=True, build_id=alt_build_id
        ) as alt_f:
            self.server.build_ids[build_id] = {
                "executable": loadable_file.name,
                "debuginfo": debug_file.name,
            }
            self.server.build_ids[alt_build_id] = {"debuginfo": alt_f.name}

            module = self.prog.extra_module("foo", create=True)
            module.build_id = build_id
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.loaded_file_path,
                str(self.cache_dir / build_id.hex() / "executable"),
            )
            self.assertEqual(
                module.debug_file_path,
                str(self.cache_dir / build_id.hex() / "debuginfo"),
            )
            self.assertEqual(
                module.supplementary_debug_file_path,
                str(self.cache_dir / alt_build_id.hex() / "debuginfo"),
            )

    def test_gnu_debugaltlink_not_found(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with NamedTemporaryElfFile(
            loadable=True, debug=False, build_id=build_id
        ) as loadable_file, NamedTemporaryElfFile(
            loadable=False,
            debug=True,
            build_id=build_id,
            gnu_debugaltlink=("alt.debug", alt_build_id),
        ) as debug_file:
            self.server.build_ids[build_id] = {
                "executable": loadable_file.name,
                "debuginfo": debug_file.name,
            }

            module = self.prog.extra_module("foo", create=True)
            module.build_id = build_id
            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertEqual(
                module.wanted_supplementary_debug_file(),
                (
                    SupplementaryFileKind.GNU_DEBUGALTLINK,
                    str(self.cache_dir / build_id.hex() / "debuginfo"),
                    "alt.debug",
                    alt_build_id,
                ),
            )
            self.assertEqual(
                module.loaded_file_path,
                str(self.cache_dir / build_id.hex() / "executable"),
            )

    def test_only_gnu_debugaltlink(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with NamedTemporaryElfFile(
            build_id=build_id,
            gnu_debugaltlink=("alt.debug", alt_build_id),
        ) as f, NamedTemporaryElfFile(
            loadable=False, debug=True, build_id=alt_build_id
        ) as alt_f:
            self.server.build_ids[alt_build_id] = {"debuginfo": alt_f.name}

            module = self.prog.extra_module("foo", create=True)
            module.try_file(f.name)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertEqual(module.loaded_file_path, f.name)

            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(module.debug_file_path, f.name)
            self.assertEqual(
                module.supplementary_debug_file_path,
                str(self.cache_dir / alt_build_id.hex() / "debuginfo"),
            )

    def test_only_gnu_debugaltlink_not_found(self):
        build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        alt_build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"

        with NamedTemporaryElfFile(
            build_id=build_id,
            gnu_debugaltlink=("alt.debug", alt_build_id),
        ) as f:
            module = self.prog.extra_module("foo", create=True)
            module.try_file(f.name)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )
            self.assertEqual(
                module.wanted_supplementary_debug_file(),
                (
                    SupplementaryFileKind.GNU_DEBUGALTLINK,
                    f.name,
                    "alt.debug",
                    alt_build_id,
                ),
            )
            self.assertEqual(module.loaded_file_path, f.name)

            self.prog.load_module_debug_info(module)
            self.assertEqual(module.loaded_file_status, ModuleFileStatus.HAVE)
            self.assertEqual(
                module.debug_file_status, ModuleFileStatus.WANT_SUPPLEMENTARY
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import gzip
import re

from drgn.helpers.linux.kconfig import get_kconfig
from tests.linux_kernel import LinuxKernelTestCase


class TestKconfig(LinuxKernelTestCase):
    def test_get_kconfig(self):
        expected = {}
        try:
            with gzip.open("/proc/config.gz", "rt") as f:
                for line in f:
                    match = re.match(r"(\w+)=(.*)", line)
                    if match:
                        expected[match.group(1)] = match.group(2)
        except FileNotFoundError:
            self.skipTest("kernel not built with CONFIG_IKCONFIG_PROC")
        self.assertEqual(get_kconfig(self.prog), expected)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse

from tests.resources import get_resource

parser = argparse.ArgumentParser(
    description="decompress test resources and print their paths"
)
parser.add_argument("name", nargs="+", help="resource name")
args = parser.parse_args()

for name in args.name:
    print(get_resource(name))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

r"""
This script reads a drgn architecture definition file ("arch_foo_defs.py") and
outputs architecture definition code which must then be processed by
gen_strswitch.py to produce the final "arch_foo.inc" file.

The definition file must define three global variables:

* REGISTERS must be a sequence of DrgnRegister defining the names of all of the
  registers in the architecture. They should be defined in the architecture's
  logical order. See DrgnRegister below for more details.

* REGISTER_LAYOUT must be a sequence of DrgnRegisterLayout defining how drgn
  stores the registers in memory. The defined registers are laid out
  sequentially. To minimize memory usage, registers can be ordered by how
  commonly they are saved (typically, the program counter/return address
  register first, then the stack pointer register, then callee-saved registers,
  and last caller-saved registers). They can also be ordered to match the
  format in, e.g., NT_PRSTATUS. Every register in REGISTERS must have an entry
  in REGISTER_LAYOUT. However, the converse does not need to be true. For
  example, REGISTER_LAYOUT can contain pseudo-registers that are not in
  REGISTERS and are thus not exposed to the API. See DrgnRegisterLayout below
  for more details.

* STACK_POINTER_REGISTER must be the identifier of the register which contains
  the stack pointer.

The generated file includes "arch_register_layout.h" and defines several things:

1. An enum of register numbers used to implement DRGN_REGISTER_NUMBER().

2. A structure corresponding to the register layout used to implement
   DRGN_REGISTER_OFFSET(), DRGN_REGISTER_SIZE(), and DRGN_REGISTER_END():
   struct drgn_arch_register_layout;

3. An array of register layouts indexed by internal register number:
   static const struct drgn_register_layout register_layout[];

4. An array of register definitions:
   static const struct drgn_register registers[];

5. A name lookup function:
   static const struct drgn_register *register_by_name(const char *name);

6. A mapping from DWARF register numbers to internal register numbers:
   static drgn_register_number dwarf_regno_to_internal(uint64_t dwarf_regno);

7. A macro containing initializers for the "register_layout",
   "dwarf_regno_to_internal", "registers", "num_registers",
   "stack_pointer_regno", and "register_by_name" members of
   "struct drgn_architecture_info":
   #define DRGN_ARCHITECTURE_REGISTERS ...
"""

import argparse
import re
import runpy
import sys
from typing import Optional, Sequence, TextIO, Union

from codegen_utils import c_string_literal


class ArchDefsError(Exception):
    pass


class DrgnRegister:
    def __init__(
        self, names: Union[str, Sequence[str]], *, identifier: Optional[str] = None
    ) -> None:
        """
        Register definition.

        :param names: Name or names of the register (i.e., name used for this
            register in assembly language or documentation).
        :param identifier: Register identifier matching
            :attr:`DrgnRegisterLayout.identifier`. Defaults to the first
            register name with non-alphanumeric characters replaced by "_" and
            prepended with "_" if the first character is numeric.
        """
        self.names = (names,) if isinstance(names, str) else names
        for name in self.names:
            if not name:
                raise ArchDefsError(f"invalid register name {name!r}")
        if identifier is None:
            identifier = re.sub(r"[^a-zA-Z0-9_]", "_", self.names[0])
            if not re.match(r"^[a-zA-Z_]", identifier):
                identifier = "_" + identifier
        self.identifier = identifier


class DrgnRegisterLayout:
    def __init__(
        self, identifier: str, *, size: int, dwarf_number: Optional[int]
    ) -> None:
        """
        Register layout definition.

        :param identifier: C identifier for the register to use in the
            generated file.
        :param size: Size used to store the register in bytes.
        :param dwarf_number: DWARF register number, or ``None`` if the register
            is not represented in DWARF.
        """
        self.identifier = identifier
        if not re.fullmatch(r"^[a-zA-Z_][a-zA-Z0-9_]*", identifier):
            raise ArchDefsError(f"invalid register identifier {identifier!r}")
        self.size = size
        if size <= 0:
            raise ArchDefsError(f"invalid register layout size {size}")
        self.dwarf_number = dwarf_number


def validate_register_defs(
    registers: Sequence[DrgnRegister],
    register_layout: Sequence[DrgnRegisterLayout],
    stack_pointer_register: str,
) -> None:
    layout_by_identifier = {}
    for layout in register_layout:
        if layout.identifier in layout_by_identifier:
            raise ArchDefsError(
                f"register identifier {layout.identifier!r} is duplicated in REGISTER_LAYOUT"
            )
        layout_by_identifier[layout.identifier] = layout

    names = set()
    register_identifiers = set()
    for register in registers:
        for name in register.names:
            if name in names:
                raise ArchDefsError(
                    f"register name {name!r} is duplicated in REGISTERS"
                )
            names.add(name)
        if register.identifier not in layout_by_identifier:
            raise ArchDefsError(
                f"register identifier {register.identifier} in REGISTERS is not in REGISTER_LAYOUT"
            )
        if register.identifier in register_identifiers:
            raise ArchDefsError(
                f"register identifier {register.identifier!r} is duplicated in REGISTERS"
            )
        register_identifiers.add(register.identifier)

    if stack_pointer_register not in layout_by_identifier:
        raise ArchDefsError(
            f"stack pointer register identifier {stack_pointer_register!r} is not in REGISTER_LAYOUT"
        )
    if layout_by_identifier[stack_pointer_register].size > 8:
        raise ArchDefsError(
            f"stack pointer register {stack_pointer_register!r} is too big"
        )


def gen_arch_inc_strswitch(
    registers: Sequence[DrgnRegister],
    register_layout: Sequence[DrgnRegisterLayout],
    stack_pointer_register: str,
    out_file: TextIO,
) -> None:
    validate_register_defs(registers, register_layout, stack_pointer_register)

    out_file.write("/* Generated by libdrgn/build-aux/gen_arch_inc_strswitch.py. */\n")

    out_file.write("\n")
    out_file.write("enum {\n")
    for layout in register_layout:
        out_file.write(f"\tDRGN_REGISTER_NUMBER__{layout.identifier},\n")
    out_file.write("};\n")

    out_file.write("\n")
    out_file.write("struct drgn_arch_register_layout {\n")
    for layout in register_layout:
        out_file.write(f"\tchar {layout.identifier}[{layout.size}];\n")
    out_file.write("};\n")

    out_file.write("\n")
    out_file.write("static const struct drgn_register_layout register_layout[] = {\n")
    for layout in register_layout:
        out_file.write(
            f"\t{{ offsetof(struct drgn_arch_register_layout, {layout.identifier}), {layout.size} }},\n"
        )
    out_file.write("};\n")

    out_file.write("\n")
    out_file.write(
        "static drgn_register_number dwarf_regno_to_internal(uint64_t dwarf_regno)\n"
    )
    out_file.write("{\n")
    out_file.write("\tswitch (dwarf_regno) {\n")
    for layout in register_layout:
        if layout.dwarf_number is not None:
            out_file.write(f"\tcase {layout.dwarf_number}:\n")
            out_file.write(f"\t\treturn DRGN_REGISTER_NUMBER__{layout.identifier};\n")
    out_file.write("\tdefault:\n")
    out_file.write("\t\treturn DRGN_REGISTER_NUMBER_UNKNOWN;\n")
    out_file.write("\t}\n")
    out_file.write("}\n")

    out_file.write("\n")
    out_file.write("static const struct drgn_register registers[] = {\n")
    for register in registers:
        out_file.write("\t{\n")
        out_file.write("\t\t.names = (const char * const []){\n")
        for name in register.names:
            out_file.write(f"\t\t\t{c_string_literal(name)},\n")
        out_file.write("\t\t},\n")
        out_file.write(f"\t\t.num_names = {len(register.names)},\n")
        out_file.write(f"\t\t.regno = DRGN_REGISTER_NUMBER__{register.identifier},\n")
        out_file.write("\t},\n")
    out_file.write("};\n")

    out_file.write("\n")
    out_file.write(
        "static const struct drgn_register *register_by_name(const char *name)\n"
    )
    out_file.write("{\n")
    out_file.write("\t@strswitch (name)@\n")
    for i, register in enumerate(registers):
        for name in register.names:
            out_file.write(f"\t@case {c_string_literal(name)}@\n")
            out_file.write(f"\t\treturn &registers[{i}];\n")
    out_file.write("\t@default@\n")
    out_file.write("\t\treturn NULL;\n")
    out_file.write("\t@endswitch@\n")
    out_file.write("}\n")

    out_file.write("\n")
    out_file.write("#define DRGN_ARCHITECTURE_REGISTERS \\\n")
    out_file.write("\t.register_layout = register_layout, \\\n")
    out_file.write("\t.dwarf_regno_to_internal = dwarf_regno_to_internal, \\\n")
    out_file.write("\t.registers = registers, \\\n")
    out_file.write(f"\t.num_registers = {len(registers)}, \\\n")
    out_file.write(
        f"\t.stack_pointer_regno = DRGN_REGISTER_NUMBER__{stack_pointer_register}, \\\n"
    )
    out_file.write("\t.register_by_name = register_by_name\n")


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Generate arch_foo.inc.strswitch from arch_foo_defs.py"
    )
    parser.add_argument("input", metavar="FILE", help="input arch_foo_defs.py file")
    args = parser.parse_args()

    try:
        defs = runpy.run_path(
            args.input,
            init_globals={
                "DrgnRegister": DrgnRegister,
                "DrgnRegisterLayout": DrgnRegisterLayout,
            },
        )
        try:
            registers = defs["REGISTERS"]
            register_layout = defs["REGISTER_LAYOUT"]
            stack_pointer_register = defs["STACK_POINTER_REGISTER"]
        except KeyError as e:
            sys.exit(f"{e.args[0]} is not defined")
        gen_arch_inc_strswitch(
            registers, register_layout, stack_pointer_register, sys.stdout
        )
    except ArchDefsError as e:
        sys.exit(e)


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from multiprocessing import Barrier, Process
import os
from threading import Condition, Thread

from drgn.helpers.linux.pid import (
    find_pid,
    find_task,
    for_each_pid,
    for_each_task,
    for_each_task_in_group,
)
from tests.linux_kernel import LinuxKernelTestCase


class TestPid(LinuxKernelTestCase):
    def test_find_pid(self):
        pid = os.getpid()
        self.assertEqual(find_pid(self.prog, pid).numbers[0].nr, pid)

    def test_for_each_pid(self):
        pid = os.getpid()
        self.assertTrue(
            any(
                pid_struct.numbers[0].nr == pid
                for pid_struct in for_each_pid(self.prog)
            )
        )

    def test_find_task(self):
        pid = os.getpid()
        with open("/proc/self/comm", "rb") as f:
            comm = f.read()[:-1]
        task = find_task(self.prog, os.getpid())
        self.assertEqual(task.pid, pid)
        self.assertEqual(task.comm.string_(), comm)

    def test_for_each_task(self):
        NUM_PROCS = 12
        barrier = Barrier(NUM_PROCS + 1)

        try:
            procs = [Process(target=barrier.wait) for _ in range(NUM_PROCS)]
            for proc in procs:
                proc.start()
            pids = {task.pid.value_() for task in for_each_task(self.prog)}
            for proc in procs:
                self.assertIn(proc.pid, pids)
            self.assertIn(os.getpid(), pids)
            self.assertNotIn(0, pids)
            barrier.wait()
        except BaseException:
            barrier.abort()
            for proc in procs:
                proc.terminate()
            raise

    def test_for_each_task_idle(self):
        self.assertTrue(
            any(task.pid.value_() == 0 for task in for_each_task(self.prog, idle=True))
        )

    def test_for_each_task_in_group(self):
        NUM_THREADS = 12
        condition = Condition()
        this_task = find_task(self.prog, os.getpid())

        def thread_func():
            with condition:
                condition.wait()

        try:
            threads = [Thread(target=thread_func) for _ in range(NUM_THREADS)]
            for thread in threads:
                thread.start()

            actual = {
                t.pid.value_()
                for t in for_each_task_in_group(this_task, include_self=False)
            }
            for thread in threads:
                self.assertIn(thread.native_id, actual)
            self.assertNotIn(os.getpid(), actual)
        finally:
            with condition:
                condition.notify_all()
            for thread in threads:
                thread.join()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import itertools
import sys
from typing import TextIO

from codegen_utils import c_string_literal

DWARF_INDEX_SECTIONS = (
    ".debug_info",
    ".debug_types",
    ".debug_abbrev",
    ".debug_str",
    ".debug_str_offsets",
)

CACHED_SECTIONS = (
    ".debug_addr",
    ".debug_frame",
    ".eh_frame",
    ".debug_loc",
    ".debug_loclists",
)

UNCACHED_SECTIONS = (
    ".text",
    ".got",
    ".gnu_debuglink",
    ".gnu_debugaltlink",
)


def section_enumerator_name(section_name: str) -> str:
    return "DRGN_SCN_" + section_name.lstrip(".").upper()


def gen_elf_sections_h(out_file: TextIO) -> None:
    out_file.write(
        """\
/* Generated by libdrgn/build-aux/gen_elf_sections.py -H. */

#ifndef DRGN_ELF_SECTIONS_H
#define DRGN_ELF_SECTIONS_H

/**
 * Identifiers for important ELF sections so that they can be referenced by
 * index rather than name.
 */
enum drgn_section_index {
"""
    )
    for section_name in DWARF_INDEX_SECTIONS:
        out_file.write(f"\t{section_enumerator_name(section_name)},\n")
    out_file.write(
        """\
	/** Indices less than this are used by the DWARF index. */
	DRGN_SECTION_INDEX_NUM_DWARF_INDEX,
"""
    )

    for i, section_name in enumerate(CACHED_SECTIONS):
        if i == 0:
            out_file.write(
                f"\t{section_enumerator_name(section_name)} = DRGN_SECTION_INDEX_NUM_DWARF_INDEX,\n"
            )
        else:
            out_file.write(f"\t{section_enumerator_name(section_name)},\n")
    out_file.write(
        """\
	/** Indices less than this may have their data cached. */
	DRGN_SECTION_INDEX_NUM_DATA,
"""
    )

    for i, section_name in enumerate(UNCACHED_SECTIONS):
        if i == 0:
            out_file.write(
                f"\t{section_enumerator_name(section_name)} = DRGN_SECTION_INDEX_NUM_DATA,\n"
            )
        else:
            out_file.write(f"\t{section_enumerator_name(section_name)},\n")
    out_file.write(
        """\
	/** Number of section indices. */
	DRGN_SECTION_INDEX_NUM
};

#endif /* DRGN_ELF_SECTIONS_H */
"""
    )


def gen_drgn_section_name_to_index_inc_strswitch(out_file: TextIO) -> None:
    debug_sections = []
    non_debug_sections = []
    for section_name in itertools.chain(
        DWARF_INDEX_SECTIONS, CACHED_SECTIONS, UNCACHED_SECTIONS
    ):
        if section_name.startswith(".debug_"):
            debug_sections.append(section_name)
        else:
            non_debug_sections.append(section_name)

    out_file.write(
        """\
/* Generated by libdrgn/build-aux/gen_elf_sections.py. */

static enum drgn_section_index drgn_debug_section_name_to_index(const char *name, size_t len)
{
	@memswitch (name, len)@
"""
    )
    for section_name in debug_sections:
        out_file.write(f"\t@case {c_string_literal(section_name[len('.debug_'):])}@\n")
        out_file.write(f"\t\treturn {section_enumerator_name(section_name)};\n")
    out_file.write(
        """\
	@default@
		return DRGN_SECTION_INDEX_NUM;
	@endswitch@
}

static enum drgn_section_index drgn_non_debug_section_name_to_index(const char *name)
{
	@strswitch (name)@
"""
    )
    for section_name in non_debug_sections:
        out_file.write(f"\t@case {c_string_literal(section_name)}@\n")
        out_file.write(f"\t\treturn {section_enumerator_name(section_name)};\n")
    out_file.write(
        """\
	@default@
		return DRGN_SECTION_INDEX_NUM;
	@endswitch@
}
"""
    )


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--header", "-H", action="store_true", help="generate header file"
    )
    args = parser.parse_args()

    if args.header:
        gen_elf_sections_h(sys.stdout)
    else:
        gen_drgn_section_name_to_index_inc_strswitch(sys.stdout)


if __name__ == "__main__":
    main()
